# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Module for handling GROMACS input/output.

Important methods defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

read_gromacs_generic (:py:func:`.read_gromacs_generic`)
    A method for reading snapshots from any GROMACS file.

read_gromacs_file (:py:func:`.read_gromacs_file`)
    A method for reading snapshots from a GROMACS GRO file.

read_gromacs_gro_file (:py:func:`.read_gromacs_gro_file`)
    Read a single snapshot from a GROMACS GRO file.

write_gromacs_gro_file (:py:func:`.write_gromacs_gro_file`)
    Write configuration in GROMACS GRO format.

read_gromos96_file (:py:func:`.read_gromos96_file`)
    Read a single configuration GROMACS .g96 file.

write_gromos96_file (:py:func:`.write_gromos96_file`)
    Write configuration in GROMACS g96 format.

read_xvg_file (:py:func:`.read_xvg_file`)
    For reading .xvg files from GROMACS.

read_trr_header (:py:func:`.read_trr_header`)
    Read a header from an open TRR file.

read_trr_data (:py:func:`.read_trr_data`)
    Read data from an open TRR file.

skip_trr_data (:py:func:`.skip_trr_data`)
    Skip reading data from an open TRR file and move on to the
    next header.

read_trr_file (:py:func:`.read_trr_file`)
    Yield frames from a TRR file.

trr_frame_to_g96 (:py:func:`.trr_frame_to_g96`)
    Dump a specific frame from a TRR file to a .g96 file.

write_trr_frame (:py:func:`.write_trr_frame`)
    A simple method to write to a TRR file.
"""
import logging
import os
import struct
import io
import numpy as np
from pyretis.core.box import box_matrix_to_list
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())

np.set_printoptions(legacy='1.21')

__all__ = [
    'read_gromacs_generic',
    'read_gromacs_file',
    'read_gromacs_gro_file',
    'write_gromacs_gro_file',
    'read_gromos96_file',
    'write_gromos96_file',
    'read_xvg_file',
    'read_trr_header',
    'read_trr_data',
    'skip_trr_data',
    'read_trr_file',
    'trr_frame_to_g96',
    'write_trr_frame',
]


# Define formats for the trajectory output:
_GRO_FMT = '{0:5d}{1:5s}{2:5s}{3:5d}{4:8.3f}{5:8.3f}{6:8.3f}'
_GRO_VEL_FMT = _GRO_FMT + '{7:8.4f}{8:8.4f}{9:8.4f}'
_GRO_BOX_FMT = '{:15.9f}'
_G96_FMT = '{0:}{1:15.9f}{2:15.9f}{3:15.9f}\n'
_G96_FMT_FULL = '{0:5d} {1:5s} {2:5s}{3:7d}{4:15.9f}{5:15.9f}{6:15.9f}\n'
_G96_BOX_FMT = '{:15.9f}' * 9 + '\n'
_G96_BOX_FMT_3 = '{:15.9f}' * 3 + '\n'

# Definitions for the TRR reader.
_GROMACS_MAGIC = 1993
_DIM = 3
_TRR_VERSION = 'GMX_trn_file'
_TRR_VERSION_B = b'GMX_trn_file'
_SIZE_FLOAT = struct.calcsize('f')
_SIZE_DOUBLE = struct.calcsize('d')
_HEAD_FMT = '{}13i'
_HEAD_ITEMS = ('ir_size', 'e_size', 'box_size', 'vir_size', 'pres_size',
               'top_size', 'sym_size', 'x_size', 'v_size', 'f_size',
               'natoms', 'step', 'nre', 'time', 'lambda')
TRR_DATA_ITEMS = ('box_size', 'vir_size', 'pres_size',
                  'x_size', 'v_size', 'f_size')


def read_gromacs_lines(lines):
    """Read and parse GROMACS GRO data.

    This method will read a GROMACS file and yield the different
    snapshots found in the file.

    Parameters
    ----------
    lines : iterable
        Some lines of text data representing a GROMACS GRO file.

    Yields
    ------
    out : dict
        This dict contains the snapshot.

    """
    lines_to_read = 0
    snapshot = {}
    read_natoms = False
    gro = (5, 5, 5, 5, 8, 8, 8, 8, 8, 8)
    gro_keys = ('residunr', 'residuname', 'atomname', 'atomnr',
                'x', 'y', 'z', 'vx', 'vy', 'vz')
    gro_type = (int, str, str, int, float, float, float, float, float, float)
    for line in lines:
        if read_natoms:
            read_natoms = False
            lines_to_read = int(line.strip()) + 1
            continue  # just skip to next line
        if lines_to_read == 0:  # new snapshot
            if snapshot:
                _add_matrices_to_snapshot(snapshot)
                yield snapshot
            snapshot = {'header': line.strip()}
            read_natoms = True
        elif lines_to_read == 1:  # read box
            snapshot['box'] = np.array(
                [float(i) for i in line.strip().split()]
            )
            lines_to_read -= 1
        else:  # read atoms
            lines_to_read -= 1
            current = 0
            for i, key, gtype in zip(gro, gro_keys, gro_type):
                val = line[current:current+i].strip()
                if not val:
                    # This typically happens if we try to read velocities
                    # and they are not present in the file.
                    break
                value = gtype(val)
                current += i
                try:
                    snapshot[key].append(value)
                except KeyError:
                    snapshot[key] = [value]
    if snapshot:
        _add_matrices_to_snapshot(snapshot)
        yield snapshot


def _add_matrices_to_snapshot(snapshot):
    """Extract positions and velocities as matrices from GROMACS.

    The extracted positions and velocities will be added to the given
    snapshot.

    Parameters
    ----------
    snapshot : dict
        This dict contains the data read from the GROMACS file.

    Returns
    -------
    xyz : numpy.array
        The positions as an array, (N, 3).
    vel : numpy.array
        The velocities as an array, (N, 3).

    """
    xyz = np.zeros((len(snapshot['atomnr']), 3))
    for i, key in enumerate(('x', 'y', 'z')):
        if key in snapshot:
            xyz[:, i] = snapshot[key]
    vel = np.zeros_like(xyz)
    for i, key in enumerate(('vx', 'vy', 'vz')):
        if key in snapshot:
            vel[:, i] = snapshot[key]
    snapshot['xyz'] = xyz
    snapshot['vel'] = vel
    return xyz, vel


def read_gromacs_generic(filename):
    """Read GROMACS files.

    This method will read a GROMACS file and yield the different
    snapshots found in the file. This file is intended to be used
    to just count the n of snapshots stored in a file.

    Parameters
    ----------
    filename : string
        The file to check.

    Yields
    ------
    out : None.

    """
    if filename[-4:] == '.gro':
        for _ in read_gromacs_file(filename):
            yield None
    if filename[-4:] == '.g96':
        yield None
    if filename[-4:] == '.trr':
        for _ in read_trr_file(filename):
            yield None


def read_gromacs_file(filename):
    """Read GROMACS GRO files.

    This method will read a GROMACS file and yield the different
    snapshots found in the file. This file is intended to be used
    if we want to read all snapshots present in a file.

    Parameters
    ----------
    filename : string
        The file to open.

    Yields
    ------
    out : dict
        This dict contains the snapshot.

    Examples
    --------
    >>> from pyretis.inout.formats.gromacs import read_gromacs_file
    >>> for snapshot in read_gromacs_file('traj.gro'):
    ...     print(snapshot['x'][0])

    """
    with open(filename, 'r', encoding='utf-8') as fileh:
        for snapshot in read_gromacs_lines(fileh):
            yield snapshot


def read_gromacs_gro_file(filename):
    """Read a single configuration GROMACS GRO file.

    This method will read the first configuration from the GROMACS
    GRO file and return the data as give by
    :py:func:`.read_gromacs_lines`. It will also explicitly
    return the matrices with positions, velocities and box size.

    Parameters
    ----------
    filename : string
        The file to read.

    Returns
    -------
    frame : dict
        This dict contains all the data read from the file.
    xyz : numpy.array
        The positions. The array is (N, 3) where N is the
        number of particles.
    vel : numpy.array
        The velocities. The array is (N, 3) where N is the
        number of particles.
    box : numpy.array
        The box dimensions.

    """
    snapshot = None
    xyz = None
    vel = None
    box = None
    with open(filename, 'r', encoding='utf8') as fileh:
        snapshot = next(read_gromacs_lines(fileh))
        box = snapshot.get('box', None)
        xyz = snapshot.get('xyz', None)
        vel = snapshot.get('vel', None)
    return snapshot, xyz, vel, box


def write_gromacs_gro_file(outfile, txt, xyz, vel=None, box=None):
    """Write configuration in GROMACS GRO format.

    Parameters
    ----------
    outfile : string
        The name of the file to create.
    txt : dict of lists of strings
        This dict contains the information on residue-numbers, names,
        etc. required to write the GRO file.
    xyz : numpy.array
        The positions to write.
    vel : numpy.array, optional
        The velocities to write.
    box: numpy.array, optional
        The box matrix.

    """
    resnum = txt['residunr']
    resname = txt['residuname']
    atomname = txt['atomname']
    atomnr = txt['atomnr']
    npart = len(xyz)
    with open(outfile, 'w', encoding='utf-8') as output:
        output.write(f'{txt["header"]}\n')
        output.write(f'{npart}\n')
        for i in range(npart):
            if vel is None:
                buff = _GRO_FMT.format(
                    resnum[i],
                    resname[i],
                    atomname[i],
                    atomnr[i],
                    xyz[i, 0],
                    xyz[i, 1],
                    xyz[i, 2])
            else:
                buff = _GRO_VEL_FMT.format(
                    resnum[i],
                    resname[i],
                    atomname[i],
                    atomnr[i],
                    xyz[i, 0],
                    xyz[i, 1],
                    xyz[i, 2],
                    vel[i, 0],
                    vel[i, 1],
                    vel[i, 2])
            output.write(f'{buff}\n')
        if box is None:
            box = ' '.join([_GRO_BOX_FMT.format(i) for i in txt['box']])
        else:
            box = ' '.join([_GRO_BOX_FMT.format(i) for i in box])
        output.write(f'{box}\n')


def read_gromos96_file(filename):
    """Read a single configuration GROMACS .g96 file.

    Parameters
    ----------
    filename : string
        The file to read.

    Returns
    -------
    rawdata : dict of list of strings
        This is the raw data read from the file grouped into sections.
        Note that this does not include the actual positions and
        velocities as these are returned separately.
    xyz : numpy.array
        The positions.
    vel : numpy.array
        The velocities.
    box : numpy.array
        The simulation box.

    """
    _len = 15
    _pos = 24
    rawdata = {'TITLE': [], 'POSITION': [], 'VELOCITY': [], 'BOX': [],
               'POSITIONRED': [], 'VELOCITYRED': []}
    section = None
    with open(filename, 'r', encoding='utf-8', errors='replace') as gromosfile:
        for lines in gromosfile:
            new_section = False
            stripline = lines.strip()
            if stripline == 'END':
                continue
            for key in rawdata:
                if stripline == key:
                    new_section = True
                    section = key
                    break
            if new_section:
                continue
            rawdata[section].append(lines.rstrip())
    txtdata = {}
    xyzdata = {}
    for key in ('POSITION', 'VELOCITY'):
        txtdata[key] = []
        xyzdata[key] = []
        for line in rawdata[key]:
            txt = line[:_pos]
            txtdata[key].append(txt)
            pos = [float(line[i:i+_len]) for i in range(_pos, 4*_len, _len)]
            xyzdata[key].append(pos)
        for line in rawdata[key+'RED']:
            txt = line[:_pos]
            txtdata[key].append(txt)
            pos = [float(line[i:i+_len]) for i in range(0, 3*_len, _len)]
            xyzdata[key].append(pos)
        xyzdata[key] = np.array(xyzdata[key])
    rawdata['POSITION'] = txtdata['POSITION']
    rawdata['VELOCITY'] = txtdata['VELOCITY']
    if not rawdata['VELOCITY']:
        # No velocities were found in the input file.
        xyzdata['VELOCITY'] = np.zeros_like(xyzdata['POSITION'])
        logger.info('Input g96 did not contain velocities')
    if rawdata['BOX']:
        box = np.array([float(i) for i in rawdata['BOX'][0].split()])
    else:
        box = None
        logger.info('Input g96 did not contain box vectors.')
    return rawdata, xyzdata['POSITION'], xyzdata['VELOCITY'], box


def write_gromos96_file(filename, raw, xyz, vel, box=None):
    """Write configuration in GROMACS .g96 format.

    Parameters
    ----------
    filename : string
        The name of the file to create.
    raw : dict of lists of strings
        This contains the raw data read from a .g96 file.
    xyz : numpy.array
        The positions to write.
    vel : numpy.array
        The velocities to write.
    box: numpy.array, optional
        The box matrix.

    """
    _keys = ('TITLE', 'POSITION', 'VELOCITY', 'BOX')
    with open(filename, 'w', encoding='utf-8') as outfile:
        for key in _keys:
            if key not in raw:
                continue
            outfile.write(f'{key}\n')
            for i, line in enumerate(raw[key]):
                if key == 'POSITION':
                    outfile.write(_G96_FMT.format(line, *xyz[i]))
                elif key == 'VELOCITY':
                    if vel is not None:
                        outfile.write(_G96_FMT.format(line, *vel[i]))
                elif box is not None and key == 'BOX':
                    if len(box) == 3:
                        outfile.write(_G96_BOX_FMT_3.format(*box))
                    else:
                        outfile.write(_G96_BOX_FMT.format(*box))
                else:
                    outfile.write(f'{line}\n')
            outfile.write('END\n')


def read_xvg_file(filename):
    """Return data in xvg file as numpy array."""
    data = []
    legends = []
    with open(filename, 'r', encoding='utf-8') as fileh:
        for lines in fileh:
            if lines.startswith('@ s') and lines.find('legend') != -1:
                legend = lines.split('legend')[-1].strip()
                legend = legend.replace('"', '')
                legends.append(legend.lower())
            else:
                if lines.startswith('#') or lines.startswith('@'):
                    pass
                else:
                    data.append([float(i) for i in lines.split()])
    data = np.array(data)
    data_dict = {'step': np.arange(tuple(data.shape)[0])}
    for i, key in enumerate(legends):
        data_dict[key] = data[:, i+1]
    return data_dict


def swap_integer(integer):
    """Convert little/big endian."""
    return (((integer << 24) & 0xff000000) | ((integer << 8) & 0x00ff0000) |
            ((integer >> 8) & 0x0000ff00) | ((integer >> 24) & 0x000000ff))


def swap_endian(endian):
    """Just swap the string for selecting big/little."""
    if endian == '>':
        return '<'
    if endian == '<':
        return '>'
    raise ValueError('Undefined swap!')


def read_struct_buff(fileh, fmt):
    """Unpack from a file handle with a given format.

    Parameters
    ----------
    fileh : file object
        The file handle to unpack from.
    fmt : string
        The format to use for unpacking.

    Returns
    -------
    out : tuple
        The unpacked elements according to the given format.

    Raises
    ------
    EOFError
        We will raise an EOFError if `fileh.read()` attempts to read
        past the end of the file.

    """
    buff = fileh.read(struct.calcsize(fmt))
    if not buff:
        raise EOFError
    return struct.unpack(fmt, buff)


def read_matrix(fileh, endian, double):
    """Read a matrix from the TRR file.

    Here, we assume that the matrix will be of
    dimensions (_DIM, _DIM).

    Parameters
    ----------
    fileh : file object
        The file handle to read from.
    endian : string
        Determines the byte order.
    double : boolean
        If true, we will assume that the numbers
        were stored in double precision.

    Returns
    -------
    mat : numpy.array
        The matrix as an array.

    """
    if double:
        fmt = f'{endian}{_DIM**2}d'
    else:
        fmt = f'{endian}{_DIM**2}f'
    read = read_struct_buff(fileh, fmt)
    mat = np.zeros((_DIM, _DIM))
    for i in range(_DIM):
        for j in range(_DIM):
            mat[i, j] = read[i * _DIM + j]
    return mat


def read_coord(fileh, endian, double, natoms):
    """Read a coordinate section from the TRR file.

    This method will read the full coordinate section from a TRR
    file. The coordinate section may be positions, velocities or
    forces.

    Parameters
    ----------
    fileh : file object
        The file handle to read from.
    endian : string
        Determines the byte order.
    double : boolean
        If true, we will assume that the numbers
        were stored in double precision.
    natoms : int
        The number of atoms we have stored coordinates for.

    Returns
    -------
    mat : numpy.array
        The coordinates as a numpy array. It will have
        ``natoms`` rows and ``_DIM`` columns.

    """
    if double:
        fmt = f'{endian}{natoms * _DIM}d'
    else:
        fmt = f'{endian}{natoms * _DIM}f'
    read = read_struct_buff(fileh, fmt)
    mat = np.array(read)
    mat.shape = (natoms, _DIM)
    return mat


def is_double(header):
    """Determine if we should use double precision.

    This method determined the precision to use when reading
    the TRR file. This is based on the header read for a given
    frame which defines the sizes of certain "fields" like the box
    or the positions. From this size, the precision can be obtained.

    Parameters
    ----------
    header : dict
        The header read from the TRR file.

    Returns
    -------
    out : boolean
        True if we should use double precision.

    """
    key_order = ('box_size', 'x_size', 'v_size', 'f_size')
    size = 0
    for key in key_order:
        if header[key] != 0:
            if key == 'box_size':
                size = int(header[key] / _DIM**2)
                break
            size = int(header[key] / (header['natoms'] * _DIM))
            break
    if size not in (_SIZE_FLOAT, _SIZE_DOUBLE):
        raise ValueError('Could not determine size!')
    return size == _SIZE_DOUBLE


def read_trr_header(fileh):
    """Read a header from a TRR file.

    Parameters
    ----------
    fileh : file object
        The file handle for the file we are reading.

    Returns
    -------
    header : dict
        The header read from the file.

    """
    start = fileh.tell()
    endian = '>'
    magic = read_struct_buff(fileh, f'{endian}1i')[0]
    if magic == _GROMACS_MAGIC:
        pass
    else:
        magic = swap_integer(magic)
        if not magic == _GROMACS_MAGIC:
            logger.critical(
                'TRR file might be inconsistent! Could find _GROMACS_MAGIC'
            )
        endian = swap_endian(endian)
    slen = read_struct_buff(fileh, f'{endian}2i')
    raw = read_struct_buff(fileh, f'{endian}{slen[0] - 1}s')
    version = raw[0].split(b'\0', 1)[0].decode('utf-8')
    if not version == _TRR_VERSION:
        raise ValueError('Unknown format')

    head_fmt = _HEAD_FMT.format(endian)
    head_s = read_struct_buff(fileh, head_fmt)
    header = {}
    for i, val in enumerate(head_s):
        key = _HEAD_ITEMS[i]
        header[key] = val
    # The next are either floats or double
    double = is_double(header)
    if double:
        fmt = f'{endian}2d'
    else:
        fmt = f'{endian}2f'
    header_r = read_struct_buff(fileh, fmt)
    header['time'] = header_r[0]
    header['lambda'] = header_r[1]
    header['endian'] = endian
    header['double'] = double
    return header, fileh.tell() - start


def skip_trr_data(fileh, header):
    """Skip coordinates/box data etc.

    This method is used when we want to skip a data section in
    the TRR file. Rather than reading the data, it will use the
    size read in the header to skip ahead to the next frame.

    Parameters
    ----------
    fileh : file object
        The file handle for the file we are reading.
    header : dict
        The header read from the TRR file.

    """
    offset = sum([header[key] for key in TRR_DATA_ITEMS])
    fileh.seek(offset, 1)


def read_trr_data(fileh, header):
    """Read box, coordinates etc. from a TRR file.

    Parameters
    ----------
    fileh : file object
        The file handle for the file we are reading.
    header : dict
        The header read from the file.

    Returns
    -------
    data : dict
        The data we read from the file. It may contain the following
        keys if the data was found in the frame:

        - ``box`` : the box matrix,
        - ``vir`` : the virial matrix,
        - ``pres`` : the pressure matrix,
        - ``x`` : the coordinates,
        - ``v`` : the velocities, and
        - ``f`` : the forces

    """
    data = {}
    endian = header['endian']
    double = header['double']
    for key in ('box', 'vir', 'pres'):
        header_key = f'{key}_size'
        if header[header_key] != 0:
            data[key] = read_matrix(fileh, endian, double)
    for key in ('x', 'v', 'f'):
        header_key = f'{key}_size'
        if header[header_key] != 0:
            data[key] = read_coord(fileh, endian, double,
                                   header['natoms'])
    return data


def read_trr_file(filename, read_data=True):
    """Yield frames from a TRR file."""
    with open(filename, 'rb') as infile:
        while True:
            try:
                header, _ = read_trr_header(infile)
                if read_data:
                    data = read_trr_data(infile, header)
                else:
                    skip_trr_data(infile, header)
                    data = None
                yield header, data
            except EOFError:
                return None, None
            except struct.error:
                logger.warning(
                    'Could not read a frame from the TRR file. Aborting!'
                )
                return None, None


def read_trr_frame(filename, index):
    """Return a given frame from a TRR file."""
    idx = 0
    with open(filename, 'rb') as infile:
        while True:
            try:
                header, _ = read_trr_header(infile)
                if idx == index:
                    data = read_trr_data(infile, header)
                    return header, data
                skip_trr_data(infile, header)
                idx += 1
                if idx > index:
                    logger.error('Frame %i not found in %s', index, filename)
                    return None, None
            except EOFError:
                return None, None


def trr_frame_to_g96(trr_file, index, outfile):
    """Dump a TRR frame to a GROMOS96 (.g96) file.

    Parameters
    ----------
    trr_file : string
        The TRR file to read from.
    index : integer
        The index for the frame to dump from ``trr_file``.
    outfile : string
        The g96 file to write.

    """
    header, data = read_trr_frame(trr_file, index)
    with open(outfile, 'w', encoding='utf-8') as output:
        output.write('TITLE\n')
        output.write(f' Dump from TRR, index = {index}, '
                     f'time = {header["time"]}\n')
        output.write('END\n')
        output.write('POSITION\n')
        for i, posi in enumerate(data['x']):
            output.write(_G96_FMT_FULL.format(i, 'DUM', 'X', i, *posi))
        output.write('END\n')
        if 'v' in data:
            output.write('VELOCITY\n')
            for i, veli in enumerate(data['v']):
                output.write(_G96_FMT_FULL.format(i, 'DUM', 'X', i, *veli))
            output.write('END\n')
        output.write('BOX\n')
        box = data['box']
        output.write(_G96_BOX_FMT.format(*box_matrix_to_list(box, full=True)))
        output.write('END\n')


def _get_chunks(start, end, size):
    """Yield chunks between start and end of the given size."""
    while start < end:
        new_size = min(size, end - start)
        start += new_size
        yield start, new_size


def reverse_trr(filename, outname, print_progress=True):
    """Reverse a GROMACS TRR file.

    Parameters
    ----------
    filename : string
        The file path to the input file.
    outname : string
        The file path to the output file.
    print_progress : boolean, optional
        This can be used to print out information about the
        progress to the screen. Might be useful if we are reading
        a large file.

    """
    # We first loop over the file until we reach the
    # last header, we skip the data and read all the headers.
    all_headers = []
    buff_size = io.DEFAULT_BUFFER_SIZE
    with open(filename, 'rb') as infile, open(outname, 'wb') as outfile:
        while True:
            try:
                header, header_size = read_trr_header(infile)
                header_loc = infile.tell()
                skip_trr_data(infile, header)
                all_headers.append((header, header_loc, header_size))
            except EOFError:
                break
        # Loop through headers in reverse and write data.
        for header, header_loc, header_size in reversed(all_headers):
            if print_progress:  # pragma: no cover
                print(f'Processing step {header["step"]} '
                      f'time {header["time"]}')
            data_size = sum([header[key] for key in TRR_DATA_ITEMS])
            start = header_loc - header_size
            infile.seek(start)
            end = start + header_size + data_size
            for _, chunk_size in _get_chunks(start, end, buff_size):
                outfile.write(infile.read(chunk_size))


def write_trr_frame(filename, data, endian=None, double=False, append=False):
    """Write data in TRR format to a file.

    Parameters
    ----------
    filename : string
        The name/path of the file to write to.
    data : dict
        The data we will write to the file.
    endian : string, optional
        Select the byte order; big-endian or little-endian. If not
        specified, the native byte order will be used.
    double : boolean, optional
        If True, we will write in double precision.
    append : boolean, optional
        If True, we will append to the file, otherwise, the file
        will be overwritten.

    """
    if append:
        mode = 'ab'
    else:
        mode = 'wb'

    if double:
        size = _SIZE_DOUBLE
        floatfmt = '{}d'
    else:
        floatfmt = '{}f'
        size = _SIZE_FLOAT

    if endian:
        floatfmt = endian + floatfmt

    header = {}
    for key in _HEAD_ITEMS:
        header[key] = 0

    header['natoms'] = data['natoms']
    header['step'] = data['step']
    header['box_size'] = size * _DIM * _DIM
    for i in ('x', 'v', 'f'):
        if i in data:
            header[f'{i}_size'] = data['natoms'] * size * _DIM
    header['endian'] = endian
    header['double'] = double
    header['time'] = data['time']
    header['lambda'] = data['lambda']

    with open(filename, mode) as outfile:
        write_trr_header(outfile, header, floatfmt, endian=endian)
        for key in TRR_DATA_ITEMS:
            if header[key] != 0:
                matrix = data[key.split('_', 1)[0]]
                fmt = floatfmt.format(matrix.size)
                outfile.write(struct.pack(fmt, *matrix.flatten()))
    return header


def write_trr_header(outfile, header, floatfmt, endian=None):
    """Write TRR header.

    Parameters
    ----------
    outfile : filehandle
        The file we can write to.
    header : dict
        The header data for the TRR file.
    floatfmt : string
        The string which gives the format for floats. It should indicate
        if we are writing for double or single precision.
    endian : string, optional
        Can be used to force endianess.

    """
    slen = (13, 12)
    fmt = ['1i', '2i', f'{slen[0] - 1}s', '13i']
    if endian:
        fmt = [endian + i for i in fmt]
    outfile.write(struct.pack(fmt[0], _GROMACS_MAGIC))
    outfile.write(struct.pack(fmt[1], *slen))
    outfile.write(struct.pack(fmt[2], _TRR_VERSION_B))
    head = [header[key] for key in _HEAD_ITEMS[:13]]
    outfile.write(struct.pack(fmt[3], *head))
    outfile.write(struct.pack(floatfmt.format(1), header['time']))
    outfile.write(struct.pack(floatfmt.format(1), header['lambda']))
    outfile.flush()


def gromacs_settings(settings, input_path):
    """Read and processes GROMACS settings.

    Parameters
    ----------
    settings : dict
        The current input settings..
    input_path : string
        The GROMACS input path

    """
    ext = settings['engine'].get('gmx_format', 'gro')
    default_files = {'conf': f'conf.{ext}',
                     'input_o': 'grompp.mdp',
                     'topology': 'topol.top',
                     'index': 'index.ndx'}
    settings['engine']['input_files'] = {}
    for key in ('conf', 'input_o', 'topology', 'index'):
        # Add input path and the input files if input is no given:
        settings['engine']['input_files'][key] = \
            settings['engine'].get(key,
                                   os.path.join(input_path,
                                                default_files[key]))



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test the MD Simulation class."""
from io import StringIO
import copy
import gzip
import logging
import math
import numpy as np
import os
import pathlib
import tempfile
import unittest
from pyretis.core.box import create_box
from pyretis.core.common import big_fat_comparer
from pyretis.core.particles import Particles
from pyretis.core.random_gen import create_random_generator
from pyretis.core.system import System
from pyretis.core.units import create_conversion_factors
from pyretis.engines.internal import VelocityVerlet, Langevin
from pyretis.forcefield import ForceField
from pyretis.forcefield.potentials import (
    PairLennardJonesCutnp,
    DoubleWell,
)
from pyretis.inout.simulationio import OutputTask
from pyretis.inout.screen import ScreenOutput
from pyretis.inout.settings import add_default_settings
from pyretis.orderparameter import PositionVelocity
from pyretis.simulation.md_simulation import (
    SimulationNVE,
    SimulationMDFlux,
)
from pyretis.tools.lattice import generate_lattice
from unittest.mock import patch
from .help import turn_on_logging, TxtWriter, TEST_SETTINGS

logging.disable(logging.CRITICAL)


HERE = os.path.abspath(os.path.dirname(__file__))


class ExtraAsserts:
    """Add additional assert method for crossing.

    This class is just for defining some additional methods which
    are useful when comparing crossing data in the unit tests.

    """

    @staticmethod
    def assert_equal_cross(cross1, cross2):
        """Assert that two lists of crossing data are equal."""
        if not len(cross1) == len(cross2):
            raise AssertionError('Length of crossing data differ!')
        for cri, crj in zip(cross1, cross2):
            for i, j in zip(cri, crj):
                if not i == j:
                    raise AssertionError(
                        f'Crossing data is not equal! {cri} != {crj}'
                        )

    def assert_equal_result(self, result1, result2, skip=None):
        """Compare result dictionaries from a simulation."""
        for key1 in result1.keys():
            if key1 not in result2:
                raise AssertionError(f'{key1} missing in result2')
        for key2 in result2.keys():
            if key2 not in result1:
                raise AssertionError(f'{key2} missing in result1')
        for key1, val1 in result1.items():
            if skip and key1 in skip:
                continue
            val2 = result2[key1]
            if len(val1) != len(val2):
                raise AssertionError(
                    f'Different no. of items for {key1}'
                )
            if key1 == 'thermo':
                self.assert_equal_thermo(val1, result2[key1])
            elif key1 == 'cycle':
                self.assert_equal_dict(val1, result2[key1])
            elif key1 == 'order':
                self.assert_equal_list(val1, result2[key1])
            else:
                raise AssertionError(
                    f'Comparison for {key1} not implemented'
                )

    @staticmethod
    def assert_equal_thermo(thermo1, thermo2):
        """Compare two thermo dictionaries."""
        for key1 in thermo1.keys():
            if key1 not in thermo2:
                raise AssertionError(f'{key1} missing in thermo2')
        for key2 in thermo2.keys():
            if key2 not in thermo1:
                raise AssertionError(f'{key2} missing in thermo1')
        for key1, val1 in thermo1.items():
            if key1 in ('press-tens', 'mom'):
                comp = np.allclose(val1, thermo2[key1])
            else:
                comp = math.isclose(val1, thermo2[key1])
            if not comp:
                raise AssertionError(f'Thermo {key1} differ!')

    @staticmethod
    def assert_equal_dict(dict1, dict2):
        """Compare two plain dicts."""
        for key1 in dict1:
            if key1 not in dict2:
                raise AssertionError(f'{key1} missing in dict2')
        for key2 in dict2:
            if key2 not in dict1:
                raise AssertionError(f'{key2} missing in dict1')
        for key1, val1 in dict1.items():
            if not math.isclose(val1, dict2[key1]):
                raise AssertionError(
                    f'Values for "{key1}" differ: {val1} != {dict2[key1]}!'
                    )

    @staticmethod
    def assert_equal_list(list1, list2):
        """Compare two plain lists."""
        if not len(list1) == len(list2):
            raise AssertionError('Lists have different length')
        for i, j in zip(list1, list2):
            if not math.isclose(i, j):
                raise AssertionError(f'Lists differ {i} != {j}')


def create_test_system():
    """Just set up and create a simple test system."""
    create_conversion_factors('lj')
    xyz, size = generate_lattice('fcc', [3, 3, 3], density=0.9)
    size = np.array(size)
    box = create_box(low=size[:, 0], high=size[:, 1])
    system = System(units='lj', box=box, temperature=2.0)
    system.particles = Particles(dim=3)
    for pos in xyz:
        system.add_particle(pos, vel=np.zeros_like(pos),
                            force=np.zeros_like(pos),
                            mass=1.0, name='Ar', ptype=0)
    gen_settings = {'distribution': 'maxwell', 'momentum': True, 'seed': 0}
    system.generate_velocities(**gen_settings)
    potentials = [
        PairLennardJonesCutnp(dim=3, shift=True, mixing='geometric'),
    ]
    parameters = [
        {0: {'sigma': 1, 'epsilon': 1, 'rcut': 2.5}},
    ]
    system.forcefield = ForceField(
        'Lennard Jones force field',
        potential=potentials,
        params=parameters,
    )
    return system


def create_test_systemflux(temperature=2.0):
    """Just set up and create a simple test system."""
    create_conversion_factors('lj')
    box = create_box(periodic=[False])
    system = System(units='lj', box=box, temperature=temperature)
    system.particles = Particles(dim=1)
    pos = np.array([-1.0])
    system.add_particle(pos, vel=np.zeros_like(pos),
                        force=np.zeros_like(pos),
                        mass=1.0, name='Ar', ptype=0)
    gen_settings = {'distribution': 'maxwell', 'momentum': False, 'seed': 0}
    system.generate_velocities(**gen_settings)
    potentials = [
        DoubleWell(a=1., b=2., c=0.0),
    ]
    parameters = [
        {'a': 1.0, 'b': 2.0, 'c': 0.0},
    ]
    system.forcefield = ForceField(
        '1D Double Well force field',
        potential=potentials,
        params=parameters,
    )
    return system


def load_energy(filename):
    """Load energy data from the given file."""
    energies = []
    with gzip.open(filename, 'rt') as infile:
        for lines in infile:
            energy = [float(i) for i in lines.strip().split()]
            energies.append(energy)
    return energies


def load_traj(filename):
    """Load trajectory data from the given file."""
    snapshots = []
    snapshot = []
    with gzip.open(filename, 'rt') as infile:
        for lines in infile:
            if lines.startswith('# Snapshot'):
                if snapshot:
                    snapshots.append(snapshot)
                snapshot = []
            else:
                posvel = [float(i) for i in lines.strip().split()]
                snapshot.append(posvel)
    if snapshot:
        snapshots.append(snapshot)
    return snapshots


class TestNVESimulation(unittest.TestCase, ExtraAsserts):
    """Run the tests for NVESimulation class."""

    def test_md_simulation(self):
        """Test that we can create the simulation object."""
        system = create_test_system()
        engine = VelocityVerlet(0.002)
        simulation = SimulationNVE({'system': system, 'engine': engine},
                                   {'settings': 'empty'}, {'steps': 10})
        thermo = load_energy(os.path.join(HERE, 'md-thermo.txt.gz'))
        traj = load_traj(os.path.join(HERE, 'md-traj.txt.gz'))
        keys = ('temp', 'vpot', 'ekin', 'etot', 'press')

        def task1(simulation):
            """Dummy task for the simulation."""
            return f"Hello there {simulation.cycle['step']:03d}"

        task = {'func': task1, 'result': 'hello', 'args': (simulation,)}
        add = simulation.add_task(task)
        self.assertTrue(add)
        writer = ScreenOutput(
            TxtWriter(
                'test',
                header={'labels': ['Step', 'Message'], 'width': [9, 20]}
            )
        )

        simulation.output_tasks = [
            OutputTask('Hello-test', ['hello'], writer, {'every': 1}),
        ]

        with patch('sys.stdout', new=StringIO()):
            for i_s, (step, thermoi, snap) in enumerate(zip(simulation.run(),
                                                            thermo,
                                                            traj)):
                if i_s == 0:
                    self.assertFalse('hello' in step)
                else:
                    self.assertTrue('hello' in step)

                for i, key in enumerate(keys):
                    self.assertAlmostEqual(step['thermo'][key], thermoi[i])
                posvel = np.array(snap)
                system = step['system']
                self.assertTrue(
                    np.allclose(posvel[:, :3], system.particles.pos)
                )
                self.assertTrue(
                    np.allclose(posvel[:, 3:], system.particles.vel)
                )
                self.assertFalse('order' in step)
        restart = simulation.restart_info()
        self.assertEqual(restart['simulation']['type'],
                         simulation.simulation_type)
        engine = Langevin(1.0, 2.0, rgen=None)
        with turn_on_logging():
            with self.assertLogs('pyretis.simulation.md_simulation',
                                 level='WARNING'):
                SimulationNVE(ensemble={'system': system,
                                        'engine': engine},
                              settings={'all': 'empty'},
                              controls={'steps': 10})

    def test_soft_exit_md(self):
        """Test the soft exit method for md simulation."""
        settings = TEST_SETTINGS.copy()
        add_default_settings(settings)
        with tempfile.TemporaryDirectory() as tempdir:
            engine = VelocityVerlet(0.002)
            system = create_test_system()
            settings['simulation']['exe_path'] = tempdir
            simulation = SimulationNVE({'system': system, 'engine': engine},
                                       settings=settings,
                                       controls={'steps': 10})

            with patch('sys.stdout', new=StringIO()) as stdout:
                print(simulation)
            self.assertEqual('NVE simulation\nNumber of steps to do:'
                             ' 10\nMD engine: Velocity Verlet MD integrator\n'
                             'Time step: 0.002', stdout.getvalue().strip())

            simulation.set_up_output(settings)
            self.assertEqual(tempdir, simulation.exe_dir)

            with patch('sys.stdout', new=StringIO()):
                for i, _ in enumerate(simulation.run()):
                    if i > simulation.restart_freq:
                        break
            files = [i.name for i in os.scandir(tempdir)]
            self.assertEqual(4, len(files))
            self.assertIn('pyretis.restart', files)
            self.assertIn('traj.xyz', files)
            self.assertIn('energy.txt', files)
            self.assertIn('thermo.txt', files)
            exit_file = os.path.join(tempdir, 'EXIT')
            pathlib.Path(exit_file).touch()
            with patch('sys.stdout', new=StringIO()) as stdout:
                for _ in enumerate(simulation.run()):
                    pass
                self.assertIn('soft exit', stdout.getvalue().strip())
            self.assertEqual(simulation.cycle['step'],
                             simulation.restart_freq + 2)

    def test_md_simulation_old(self):
        """Test that we can create the simulation object."""
        system = create_test_system()
        engine = VelocityVerlet(0.002)
        simulation = SimulationNVE(ensemble={'system': system,
                                             'engine': engine},
                                   settings={'all': 'empty'},
                                   controls={'steps': 10})
        for _ in simulation.run():
            pass
        self.assertTrue(
            np.allclose(system.particles.pos[0],
                        np.array([0.04907249, 0.00866652, 0.03147752]))
        )
        restart = simulation.restart_info()
        self.assertEqual(restart['simulation']['type'],
                         simulation.simulation_type)
        engine = Langevin(1.0, 2.0, rgen=None)
        with turn_on_logging():
            with self.assertLogs('pyretis.simulation.md_simulation',
                                 level='WARNING'):
                SimulationNVE(ensemble={'system': system,
                                        'engine': engine},
                              settings={'all': 'empty'},
                              controls={'steps': 10})

    def test_md_simulation_restart(self):
        """Test that we can continue with the simulation object."""
        system1 = create_test_system()
        system2 = create_test_system()
        engine1 = VelocityVerlet(0.002)
        engine2 = VelocityVerlet(0.002)
        simulation1 = SimulationNVE(
            ensemble={'system': system1,
                      'engine': engine1},
            settings={'all': 'empty'},
            controls={'steps': 5})
        simulation2 = SimulationNVE(
            ensemble={'system': system2,
                      'engine': engine2},
            settings={'all': 'empty'},
            controls={'steps': 50})
        for _ in simulation1.run():
            pass

        restart1 = simulation1.restart_info()

        simulation1.cycle['steps'] = 50
        simulation1.cycle['endcycle'] = 50

        simulation2.load_restart_info(copy.deepcopy(restart1))

        simulation2.cycle['steps'] = 50
        simulation2.cycle['endcycle'] = 50

        for _ in simulation1.run():
            pass
        for _ in simulation2.run():
            pass

        info_parts = simulation1.restart_info()
        info_single = simulation2.restart_info()
        info_parts['system'] = simulation1.system.restart_info()
        info_single['system'] = simulation2.system.restart_info()

        self.assertTrue(np.allclose(info_parts['system']['box']['length'],
                                    info_single['system']['box']['length']))
        self.assertTrue(np.allclose(info_parts['system']['particles']['pos'],
                                    info_single['system']['particles']['pos']))
        self.assertTrue(np.allclose(info_parts['system']['particles']['vel'],
                                    info_single['system']['particles']['vel']))
        self.assertEqual(info_parts['simulation']['cycle']['steps'],
                         info_single['simulation']['cycle']['steps'])
        self.assertEqual(info_parts['system']['particles']['vpot'],
                         info_single['system']['particles']['vpot'])

        simulation1 = SimulationNVE(
            ensemble={'system': system1,
                      'engine': engine1},
            settings={'all': 'empty'},
            controls={'steps': 51})

        for _ in simulation1.run():
            pass

        info_parts = simulation1.restart_info()
        info_parts['system'] = simulation1.system.restart_info()

        self.assertFalse(
            np.allclose(info_parts['system']['particles']['pos'],
                        info_single['system']['particles']['pos']))
        self.assertFalse(
            np.allclose(info_parts['system']['particles']['vel'],
                        info_single['system']['particles']['vel']))
        self.assertNotEqual(info_parts['simulation']['cycle']['steps'],
                            info_single['simulation']['cycle']['steps'])
        self.assertNotEqual(info_parts['system']['particles']['vpot'],
                            info_single['system']['particles']['vpot'])

    def test_with_order(self):
        """Test that we can add order parameter to the simulation."""
        system = create_test_system()
        engine = VelocityVerlet(0.002)
        order = PositionVelocity(0, dim='x', periodic=False)
        simulation = SimulationNVE(
            ensemble={'system': system,
                      'order_function': order,
                      'engine': engine},
            settings={'all': 'empty'},
            controls={'steps': 3})
        correct = [
            [0.0, 2.4659841156924998],
            [0.0049319682313849998, 2.4655954880460191],
            [0.009862381952184078, 2.4644409982835884],
            [0.014789732224519354, 2.4625397443524371],
        ]
        for i, step in enumerate(simulation.run()):
            self.assertTrue('order' in step)
            self.assert_equal_list(correct[i], step['order'])

    def test_step(self):
        """Test the step method wrt. the run method."""
        system = create_test_system()
        engine = VelocityVerlet(0.002)
        order = PositionVelocity(0, dim='x', periodic=False)
        simulation = SimulationNVE(
            ensemble={'system': system,
                      'order_function': order,
                      'engine': engine},
            settings={'simulation': {}},
            controls={'steps': 3})
        results1 = []
        for step in simulation.run():
            results1.append(step)

        system = create_test_system()
        engine = VelocityVerlet(0.002)
        order = PositionVelocity(0, dim='x', periodic=False)
        simulation = SimulationNVE(
            ensemble={'system': system,
                      'order_function': order,
                      'engine': engine},
            settings={'simulation': {}},
            controls={'steps': 3})
        results2 = []
        while not simulation.is_finished():
            result = simulation.step()
            results2.append(result)

        self.assertEqual(len(results1), len(results2))
        for resulti, resultj in zip(results1, results2):
            self.assert_equal_result(resulti, resultj, skip=('system',))


class TestMDFluxSimulation(unittest.TestCase, ExtraAsserts):
    """Run the tests for MD Flux Simulation class."""

    def test_md_simulation(self):
        """Test that we can create the simulation from scratch & restart."""
        system = create_test_systemflux()
        rgen = create_random_generator({'seed': 0})
        engine = Langevin(0.002, 0.3, rgen)
        order = PositionVelocity(0, dim='x', periodic=False)
        interfaces = [-0.9]
        simulation = SimulationMDFlux(
            ensemble={'system': system,
                      'order_function': order,
                      'engine': engine},
            settings={'simulation': {'interfaces': interfaces}},
            controls={'steps': 10})
        for _ in simulation.run():
            pass
        self.assertTrue(simulation.leftside_prev[0])
        # Test loading of restart info:
        restart = simulation.restart_info()
        simulation2 = SimulationMDFlux(
            ensemble={'system': system,
                      'order_function': order,
                      'engine': engine},
            settings={'simulation': {'interfaces': interfaces}},
            controls={'steps': 100})
        self.assertEqual(simulation2.cycle['step'], 0)
        simulation2.load_restart_info(restart)
        self.assertEqual(simulation2.cycle['step'], 10)

    def test_flux_results(self):
        """Test the md-flux results."""
        system = create_test_systemflux(temperature=0.05)
        engine = Langevin(0.002, 0.3)
        order = PositionVelocity(0, dim='x', periodic=False)
        interfaces = [-0.99]
        simulation = SimulationMDFlux(
            ensemble={'system': system,
                      'order_function': order,
                      'engine': engine},
            settings={'simulation': {'interfaces': interfaces}},
            controls={'endcycle': 1000})

        correct = {
            19: [(19, 0, '+')],
            374: [(374, 0, '-')],
            855: [(855, 0, '+')],
        }

        for step in simulation.run():
            if step['cycle']['steps'] in correct:
                self.assert_equal_cross(correct[step['cycle']['steps']],
                                        step['cross'])

    def test_soft_exit_md_flux(self):
        """Test the soft exit method for MDflux method."""
        settings = TEST_SETTINGS.copy()
        add_default_settings(settings)
        with tempfile.TemporaryDirectory() as tempdir:
            system = create_test_systemflux()
            settings['simulation']['exe_path'] = tempdir
            rgen = create_random_generator({'seed': 0})
            engine = Langevin(0.002, 0.3, rgen)
            order = PositionVelocity(0, dim='x', periodic=False)
            interfaces = [-0.9]
            simulation = SimulationMDFlux(
                ensemble={'system': system,
                          'order_function': order,
                          'engine': engine},
                settings={'simulation': {'interfaces': interfaces,
                                         'exe_path': tempdir}},
                controls={'steps': 10})

            with patch('sys.stdout', new=StringIO()) as stdout:
                print(simulation)
            self.assertEqual('MD-flux simulation\nNumber of steps to do: 10'
                             '\nMolecular Dynamics engine: Langevin MD '
                             'integrator\nTime step: 0.002',
                             stdout.getvalue().strip())

            simulation.set_up_output(settings)
            self.assertEqual(tempdir, simulation.exe_dir)
            with patch('sys.stdout', new=StringIO()):
                for i, _ in enumerate(simulation.run()):
                    if i > simulation.restart_freq:
                        break
            files = [i.name for i in os.scandir(tempdir)]
            self.assertEqual(5, len(files))
            self.assertIn('pyretis.restart', files)
            self.assertIn('order.txt', files)
            self.assertIn('traj.xyz', files)
            self.assertIn('cross.txt', files)
            self.assertIn('energy.txt', files)

            exit_file = os.path.join(tempdir, 'EXIT')
            pathlib.Path(exit_file).touch()
            with patch('sys.stdout', new=StringIO()) as stdout:
                for _ in enumerate(simulation.run()):
                    pass
                self.assertIn('soft exit', stdout.getvalue().strip())
            self.assertEqual(simulation.cycle['step'],
                             simulation.restart_freq + 2)

    def test_flux_restart(self):
        """Test md-flux when we are restarting."""
        system = create_test_systemflux(temperature=0.05)
        engine = Langevin(0.002, 0.3)
        order = PositionVelocity(0, dim='x', periodic=False)
        interfaces = [-0.99]
        correct = {
            19: [(19, 0, '+')],
            374: [(374, 0, '-')],
        }
        simulation1 = SimulationMDFlux(
            ensemble={'system': system,
                      'order_function': order,
                      'engine': engine},
            settings={'simulation': {'interfaces': interfaces}},
            controls={'steps': 18})
        # Test 1: We stop before a crossing
        for step in simulation1.run():
            if step['cycle']['steps'] in correct:
                self.assert_equal_cross(correct[step['cycle']['step']],
                                        step['cross'])
        restart1 = simulation1.restart_info()
        simulation2 = SimulationMDFlux(
            ensemble={'system': system,
                      'order_function': order,
                      'engine': engine},
            settings={'simulation': {'interfaces': interfaces}},
            controls={'steps': 356})
        simulation2.load_restart_info(restart1)
        for step in simulation2.run():
            if step['cycle']['steps'] in correct:
                self.assert_equal_cross(correct[step['cycle']['steps']],
                                        step['cross'])
        restart2 = simulation2.restart_info()
        simulation3 = SimulationMDFlux(
            ensemble={'system': system,
                      'order_function': order,
                      'engine': engine},
            settings={'simulation': {'interfaces': interfaces}},
            controls={'steps': 1})
        simulation3.load_restart_info(restart2)
        for step in simulation3.run():
            if step['cycle']['steps'] in correct:
                self.assert_equal_cross(correct[step['cycle']['steps']],
                                        step['cross'])

    def test_flux_simulation_restart(self):
        """Test that we can continue with the simulation object."""
        system1 = create_test_systemflux()
        system2 = create_test_systemflux()
        rgen1 = create_random_generator({'seed': 0})
        rgen2 = create_random_generator({'seed': 16})
        engine1 = Langevin(0.002, 0.3, rgen1)
        engine2 = Langevin(0.002, 0.3, rgen2)
        order1 = PositionVelocity(0, dim='x', periodic=False)
        order2 = PositionVelocity(0, dim='x', periodic=False)
        interfaces = [-0.9]
        simulation1 = SimulationMDFlux(
            ensemble={'system': system1,
                      'order_function': order1,
                      'rgen': rgen1,
                      'engine': engine1},
            settings={'simulation': {'interfaces': interfaces}},
            controls={'endcycle': 3})
        simulation2 = SimulationMDFlux(
            ensemble={'system': system2,
                      'order_function': order2,
                      'rgen': rgen2,
                      'engine': engine2},
            settings={'simulation': {'interfaces': interfaces}},
            controls={'endcycle': 8})

        for _ in simulation1.run():
            pass

        restart1 = copy.deepcopy(simulation1.restart_info())
        restart_sys1 = copy.deepcopy(simulation1.system.restart_info())

        simulation2.load_restart_info(copy.deepcopy(restart1))
        simulation2.system.load_restart_info(copy.deepcopy(restart_sys1))

        restart2 = simulation2.restart_info()
        restart_sys2 = simulation2.system.restart_info()

        self.assertTrue(big_fat_comparer(restart_sys1,
                                         restart_sys2, hard=True))
        restart1['engine']['param_iner'] = restart2['engine']['param_iner']
        restart1['engine']['init_params'] = restart2['engine']['init_params']
        restart1['simulation']['cycle'] = restart2['simulation']['cycle']
        self.assertTrue(big_fat_comparer(restart1, restart2, hard=True))

        simulation1.cycle['endcycle'] = 8
        simulation2.cycle['endcycle'] = 8
        simulation1.first_step = True

        for _ in simulation1.run():
            pass

        info_parts = simulation1.restart_info()
        info_parts['system'] = simulation1.system.restart_info()

        for _ in simulation2.run():
            pass

        pp1 = simulation1.system.particles.restart_info()
        pp2 = simulation2.system.particles.restart_info()

        self.assertTrue(big_fat_comparer(pp1, pp2, hard=True))
        info_single = simulation2.restart_info()
        info_single['system'] = simulation2.system.restart_info()

        self.assertTrue(big_fat_comparer(info_parts, info_single, hard=True))

        self.assertEqual(info_parts['simulation']['cycle']['steps'],
                         info_single['simulation']['cycle']['steps'])
        self.assertTrue(np.allclose(info_parts['system']['box']['length'],
                                    info_single['system']['box']['length']))
        self.assertTrue(np.allclose(info_parts['system']['particles']['pos'],
                                    info_single['system']['particles']['pos']))
        self.assertEqual(info_parts['system']['particles']['vpot'],
                         info_single['system']['particles']['vpot'])
        self.assertTrue(np.allclose(info_parts['system']['particles']['vel'],
                                    info_single['system']['particles']['vel']))

        simulation1 = SimulationMDFlux(
            ensemble={'system': system1,
                      'order_function': order1,
                      'rgen': rgen1,
                      'engine': engine1},
            settings={'simulation': {'interfaces': interfaces}},
            controls={'endcycle': 18})

        for _ in simulation1.run():
            pass

        info_parts = copy.deepcopy(simulation1.restart_info())
        info_parts['system'] = copy.deepcopy(simulation1.system.restart_info())

        self.assertFalse(
            np.allclose(info_parts['system']['particles']['pos'],
                        info_single['system']['particles']['pos']))
        self.assertFalse(
            np.allclose(info_parts['system']['particles']['vel'],
                        info_single['system']['particles']['vel']))
        self.assertNotEqual(info_parts['simulation']['cycle']['endcycle'],
                            info_single['simulation']['cycle']['endcycle'])
        self.assertNotEqual(info_parts['system']['particles']['vpot'],
                            info_single['system']['particles']['vpot'])


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test the GromacsEngine class."""
import logging
import os
import tempfile
import shutil
import struct
import unittest
from pyretis.core.path import Path
from pyretis.engines.gromacs2 import (GromacsEngine2,
                                      GromacsRunner,
                                      reopen_file,
                                      read_remaining_trr)
from pyretis.inout.common import make_dirs
from pyretis.orderparameter.orderparameter import Position
from .test_helpers.test_helpers import make_test_system


logging.disable(logging.CRITICAL)
HERE = os.path.abspath(os.path.dirname(__file__))
GMX_DIR = os.path.join(HERE, 'gmx_input2')
GMX = os.path.join(HERE, 'mockgmx.py')
MDRUN = os.path.join(HERE, 'mockmdrun2.py')


class GromacsEngineTest(unittest.TestCase):
    """Run the tests for the GromacsEngine."""

    def test_init(self):
        """Test the initiation."""
        eng = GromacsEngine2(gmx='echo',
                             mdrun='echo',
                             input_path=GMX_DIR,
                             timestep=0.002,
                             subcycles=10,
                             maxwarn=10,
                             gmx_format='g96',
                             write_vel=True,
                             write_force=False)
        eng.exe_dir = GMX_DIR
        with self.assertRaises(ValueError):
            GromacsEngine2(gmx='echo',
                           mdrun='echo',
                           input_path='gmx_input',
                           timestep=0.002,
                           subcycles=10,
                           gmx_format='not-a-format')
        with self.assertRaises(FileNotFoundError):
            GromacsEngine2(gmx='echo',
                           mdrun='echo',
                           input_path='missing-files',
                           timestep=0.002,
                           subcycles=10)

    def test_propagate_forward(self):
        """Test the propagate method, forward direction.

        Here, the .trr file will be created before starting the
        engine. The engine will then be lagging behind and this
        should trigger the `read_remaining_trr` method.
        """
        with tempfile.TemporaryDirectory() as tempdir:
            eng = GromacsEngine2(GMX, MDRUN, GMX_DIR, 0.002, 7,
                                 maxwarn=1, gmx_format='g96',
                                 write_vel=True,
                                 write_force=False)
            rundir = os.path.join(tempdir, 'generate2gmxf')
            # Create the directory for running:
            make_dirs(rundir)
            eng.exe_dir = rundir
            # Create the system:
            system = make_test_system((eng.input_files['conf'], 0))
            # Propagate:
            orderp = Position(0, dim='x', periodic=False)
            path = Path(None, maxlen=8)
            ensemble = {'system': system, 'order_function': orderp,
                        'interfaces': [-0.45, 10.0, 14.0]}
            success, _ = eng.propagate(path, ensemble, reverse=False)
            self.assertTrue(success)
            self.assertEqual(path.length, 4)
            initial_x = -0.422
            for i, point in enumerate(path.phasepoints):
                self.assertAlmostEqual(point.particles.ekin, eng.subcycles * i)
                self.assertAlmostEqual(point.particles.vpot,
                                       eng.subcycles * -1.0 * i)
                self.assertAlmostEqual(point.order[0],
                                       i * eng.subcycles + initial_x, places=3)
                self.assertFalse(point.particles.get_vel())

            eng.clean_up()

    def test_propagate_backward(self):
        """Test the propagate method, backward direction.

        Here, the .trr file will be created before starting the
        engine. The engine will then be lagging behind and this
        should trigger the `read_remaining_trr` method.
        """
        with tempfile.TemporaryDirectory() as tempdir:
            eng = GromacsEngine2(GMX, MDRUN, GMX_DIR, 0.002, 7,
                                 maxwarn=1, gmx_format='g96',
                                 write_vel=True,
                                 write_force=False)
            rundir = os.path.join(tempdir, 'generate2gmxb')
            # Create the directory for running:
            make_dirs(rundir)
            eng.exe_dir = rundir
            # Create the system:
            system = make_test_system((eng.input_files['conf'], 0))
            # Propagate:
            orderp = Position(0, dim='x', periodic=False)
            path = Path(None, maxlen=8)
            ensemble = {'system': system, 'order_function': orderp,
                        'interfaces': [-20., 10.0, 14.0]}
            success, _ = eng.propagate(path, ensemble, reverse=True)
            self.assertTrue(success)
            self.assertEqual(path.length, 4)
            initial_x = -0.422
            for i, point in enumerate(path.phasepoints):
                self.assertAlmostEqual(point.particles.ekin, eng.subcycles * i)
                self.assertAlmostEqual(point.particles.vpot,
                                       eng.subcycles * -1.0 * i)
                self.assertAlmostEqual(point.order[0],
                                       initial_x - i * eng.subcycles, places=3)
                self.assertTrue(point.particles.get_vel())
            eng.clean_up()

    def test_propagate_crash(self):
        """Test the propagate method when engine crashes."""
        with tempfile.TemporaryDirectory() as tempdir:
            mdrun = f'{MDRUN} -crash'
            eng = GromacsEngine2(GMX, mdrun, GMX_DIR,
                                 timestep=0.002,
                                 subcycles=7,
                                 maxwarn=1, gmx_format='g96',
                                 write_vel=True,
                                 write_force=False)
            rundir = os.path.join(tempdir, 'generate3gmxf')
            # Create the directory for running:
            make_dirs(rundir)
            eng.exe_dir = rundir
            # Create the system:
            system = make_test_system((eng.input_files['conf'], 0))
            # Propagate:
            order_function = Position(0, dim='x', periodic=False)
            path = Path(None, maxlen=8)
            with self.assertRaises(RuntimeError):
                ensemble = {'system': system, 'order_function': order_function,
                            'interfaces': [-0.45, 10.0, 14.0]}
                eng.propagate(path, ensemble, reverse=False)
            # Check the error - output:
            with open(os.path.join(rundir, 'stderr.txt'), 'r',
                      encoding="utf8") as infile:
                data = infile.readlines()
                # Following assert should be 1, but is 2 while gromacs fixes
                # their openmm imports (which raises an extra warning atm)
                # Following assert should be 1, but is 2 while gromacs fixes
                # their openmm imports (which raises an extra warning atm)
                self.assertLessEqual(len(data), 2)
                self.assertEqual(data[-1].strip(), 'Crash error for testing.')
            eng.clean_up()

    def test_propagate_sleep(self):
        """Test the propagate method.

        Here, we try to write the trr file a bit slower so that
        the class running the GROMACS simulation will have to wait
        for the data to be written.
        """
        with tempfile.TemporaryDirectory() as tempdir:
            mdrun = f'{MDRUN} -sleep'
            eng = GromacsEngine2(GMX, mdrun, GMX_DIR, 0.002, 7,
                                 maxwarn=1, gmx_format='g96',
                                 write_vel=True,
                                 write_force=False)
            rundir = os.path.join(tempdir, 'generate4gmxf')
            # Create the directory for running:
            make_dirs(rundir)
            eng.exe_dir = rundir
            # Create the system:
            system = make_test_system((eng.input_files['conf'], 0))
            # Propagate:
            order_function = Position(0, dim='x', periodic=False)
            path = Path(None, maxlen=8)
            ensemble = {'system': system, 'order_function': order_function,
                        'interfaces': [-0.45, 10.0, 14.0]}
            success, _ = eng.propagate(path, ensemble, reverse=False)
            self.assertTrue(success)
            self.assertEqual(path.length, 4)
            initial_x = -0.422
            for i, point in enumerate(path.phasepoints):
                self.assertAlmostEqual(point.particles.ekin, eng.subcycles * i)
                self.assertAlmostEqual(point.particles.vpot,
                                       eng.subcycles * -1.0 * i)
                self.assertAlmostEqual(point.order[0],
                                       i * eng.subcycles + initial_x, places=3)
            eng.clean_up()

    def test_integrate(self):
        """Test the integrate method."""
        with tempfile.TemporaryDirectory() as tempdir:
            eng = GromacsEngine2(GMX, MDRUN, GMX_DIR, 0.002, 3,
                                 maxwarn=1, gmx_format='g96',
                                 write_vel=True,
                                 write_force=False)
            rundir = os.path.join(tempdir, 'gmxintegrate2')
            # Create the directory for running:
            make_dirs(rundir)
            eng.exe_dir = rundir
            # Create the system:
            system = make_test_system((eng.input_files['conf'], 0))
            # Propagate:
            order_function = Position(0, dim='x', periodic=False)
            i = 0
            initial_x = -0.422
            ensemble = {'system': system, 'order_function': order_function}
            for step in eng.integrate(ensemble, 5):
                j = max(i - 1, 0)
                self.assertAlmostEqual(step['order'][0],
                                       j * eng.subcycles + initial_x, places=3)
                i += 1
            eng.clean_up()
            ensemble = {'system': system}
            correct = [11.578, 14.478]
            for i_s, step in enumerate(eng.integrate(ensemble, 1)):
                j = max(i - 1, 0)
                self.assertAlmostEqual(correct[i_s],
                                       j * eng.subcycles + initial_x, places=3)
                i += 1
            eng.clean_up()


class GromacsRunnerTest(unittest.TestCase):
    """Test the Runner."""
    def test_init(self):
        """Test that we can initiate the runner."""
        eng = GromacsRunner(cmd='echo',
                            trr_file='No thank you',
                            edr_file='Please dont insist',
                            exe_dir='Get lost already')
        with self.assertRaises(RuntimeError) as err:
            eng.check_poll()
        self.assertIn('GROMACS is not running.', str(err.exception))
        del eng

    def test_read_remaining_trr(self):
        """Test read_remaining_trr work with corrupted trr."""
        filepath = os.path.dirname(os.path.abspath(__file__))
        ff_trr = os.path.join(filepath, '../inout/traj.trr')

        with tempfile.TemporaryDirectory() as tempdir:
            new_file = os.path.join(tempdir, 'life.trr')
            # Empty file first
            with open(new_file, 'w', encoding="utf8"):
                pass
            with open(new_file, 'rb') as now:
                read_remaining_trr(new_file, now, 0)
            shutil.copy(ff_trr, new_file)

            # Inconsistent file after
            with open(new_file, 'ab') as now:
                now.write(b'\x00\x00\x00\x00\x00\x00\x00\x00')
            with open(new_file, 'rb') as now:
                with self.assertRaises(struct.error) as err:
                    for _ in read_remaining_trr(new_file, now, 0):
                        read_remaining_trr(new_file, now, 0)
            self.assertIn('unpack requires a buffer of 8 bytes',
                          str(err.exception))

    def test_reopen_file(self):
        """Reopen a file if the inode has changed test."""
        filepath = os.path.dirname(os.path.abspath(__file__))
        ff_temp = os.path.join(filepath, 'config.xyz')
        f_temp = tempfile.TemporaryFile(mode='w+')
        inode = os.stat(ff_temp).st_ino
        out = reopen_file(filename=ff_temp, fileh=None,
                          inode=inode, bytes_read=None)

        self.assertEqual(out, (None, None))

        out = reopen_file(filename=ff_temp, fileh=f_temp,
                          inode=16, bytes_read=0)
        out[0].close()
        self.assertNotEqual(out, (None, None))


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Animation of umbrella sampling."""
from matplotlib import pyplot as plt
from matplotlib import gridspec
import numpy as np
import colorama
from tqdm import tqdm, trange
from pyretis.inout import print_to_screen
from pyretis.setup import create_simulation, create_force_field
from pyretis.core import System, Particles
from pyretis.analysis.histogram import histogram, match_all_histograms


UMBRELLA_WINDOWS = [
    [-1.0, -0.4],
    [-0.5, -0.2],
    [-0.3, 0.0],
    [-0.1, 0.2],
    [0.1, 0.4],
    [0.3, 0.6],
    [0.5, 1.0]
]

DOUBLE_WELL = {
    'class': 'DoubleWell',
    'parameter': {'a': 1.0, 'b': 1.0, 'c': 0.02},
}

DEFAULT_SETTINGS = {
    'system': {
        'dimensions': 1,
        'units': 'metal',
        'temperature': 500,
    },
    'simulation': {
        'task': 'umbrellawindow',
        'rgen': 1,
        'mincycle': 10000,
        'maxdx': 0.1,
        'overlap': 1.0,
        'umbrella': 1,
    },
    'particles': {
        'position': {'input_file': 'position.xyz'},
        'name': ['X'],
        'ptype': [0],
    },
}

BINS = 100
HISTLIM = (-1.2, 1.2)
FIG_FREQ = 100  # How often we should store figures.
XLIM = (-1.2, 1.2)
SCALE_STEPS = 12  # Steps for animating the scaling.
XLIM_POT = (-1.2, 1.2)
YLIM_POT = (-0.3, 0.05)
BAR_FMT = {
    'window': (
        'Plotting window {n_fmt} of '
        '{total_fmt}|{bar}| [{elapsed}<{remaining}, {rate_fmt}]'
    ),
    'scale': (
        'Scale step {n_fmt}/{total_fmt}|{bar}| '
        '[{elapsed}<{remaining}, {rate_fmt}]'
    ),
    'pos': (
        'Step no. {n_fmt}/{total_fmt}|{bar}| '
        '[{elapsed}<{remaining}, {rate_fmt}]'
    ),
    'window-scale': (
        'Re-scaling window {n_fmt} of {total_fmt}|{bar}| '
        '[{elapsed}<{remaining}, {rate_fmt}]'
    ),
}


def run_umbrella_simulation(window, overlap):
    """Run a single umbrella simulation.

    Paramters
    ---------
    window : list of floats
        The (left, right) positions defining the umbrella window.
    overlap : float
        The position that must be crossed for the umbrella simulation
        to end.

    Returns
    -------
    pos : numpy.array
        The accepted positions of the particle.
    trial : numpy.array
        The positions of all trial moves.
    success : list of boolean
        True if a move was accepted.
    ener : numpy.array
        The potential energy of the system.

    """
    # Move the bias according to the umbrella:
    settings = DEFAULT_SETTINGS.copy()
    settings['simulation']['overlap'] = overlap
    settings['potential'] = [
        DOUBLE_WELL,
        {
            'class': 'RectangularWell',
            'parameter': {'left': window[0], 'right': window[1]},
        }
    ]
    # Create the umbrella simulation:
    simulation = create_simulation(settings)
    simulation.system.particles.pos = np.array([window[0]])
    simulation.system.potential()  # Recalculate potential energy.
    pos, trial, ener = [], [], []
    success = []
    for result in tqdm(simulation.run()):
        pos.append(simulation.system.particles.pos)
        trial.append(result['displace_step'][2])
        success.append(result['displace_step'][3])
        ener.append(simulation.system.particles.vpot)
    nstep = simulation.cycle['step'] - simulation.cycle['startcycle']
    print_to_screen(f'Done. Cycles: {nstep}', level='success')
    return np.array(pos), np.array(trial), success, np.array(ener)


def run_simulation():
    """Run the simulation (all umbrellas)."""
    numb = len(UMBRELLA_WINDOWS)
    trajectory, energy = [], []  # To store all trajectories & energies.
    msg = '\nRunning umbrella no: {} of {}. Location: {}'
    # We run all the umbrella simulations by looping over
    # the different umbrellas we defined:
    print_to_screen('Starting simulations:', level='info')
    for i, window in enumerate(UMBRELLA_WINDOWS):
        print_to_screen(msg.format(i + 1, numb, window))
        # Get position that must be crossed:
        over = UMBRELLA_WINDOWS[min(i + 1, numb - 1)][0]
        pos, trial, success, ener = run_umbrella_simulation(
            window,
            over,
        )
        trajectory.append([pos, trial, success])
        energy.append(ener)
    print_to_screen('Data collection done!', level='info')
    return trajectory, energy


def evaluate_potential(system, forcefield, positions):
    """Evaluate the potential energy.

    Parameters
    ----------
    system : object like :class:`.System`
        The system containing the particle position.
    forcefield : object like :class:`.ForceField`
        The force field with potential functions.
    positions : numpy.array
        The positions for which we will evaluate the potential.

    Returns
    -------
    out : numpy.array
        The potential energy

    """
    vpot = []
    for pos in positions:
        system.particles.pos = pos
        vpot.append(forcefield.evaluate_potential(system))
    return np.array(vpot)


def add_potential_plot(axes, system, forcefield):
    """Add the potential plot to the axes."""
    pos = np.linspace(-2, 2, 250)
    vpot = evaluate_potential(system, forcefield, pos)
    line_pot, = axes.plot(pos, vpot, lw=3)
    axv = axes.axvspan(xmin=-10, xmax=-10, alpha=0.1, color='#262626')
    scatter = axes.scatter(None, None, s=150, alpha=0.6, marker='o')
    scatter_trial = axes.scatter(None, None, s=150, c='black', alpha=0.6,
                                 marker='x')
    plot_obj = {
        'potential': line_pot,
        'window': axv,
        'scatter': scatter,
        'scatter_trial': scatter_trial,
    }
    return plot_obj


def add_histogram_plots(axes_hist, axes_all_hist, axes_hist_log):
    """Add histogram plot to the given axes."""
    counts = []
    probs = []
    histlog = []
    # Create empty histogram to get bins etc:
    for _ in range(len(UMBRELLA_WINDOWS)):
        hist, _, bin_mid = histogram([0], bins=BINS, limits=HISTLIM)
        delta_bin = bin_mid[1] - bin_mid[0]
        for axi, store in zip((axes_hist, axes_all_hist, axes_hist_log),
                              (counts, probs, histlog)):
            rects = axi.bar(bin_mid - 0.5 * delta_bin, hist, delta_bin,
                            alpha=0.5)
            store.append(rects)
            for rec in rects:
                rec.set_visible(False)
    plot_obj = {
        'counts': counts,
        'prob': probs,
        'histlog': histlog,
    }
    return plot_obj


def create_plots(system, forcefield):
    """Set up for plotting."""
    fig = plt.figure()
    grid = gridspec.GridSpec(2, 2)

    ax_pot = fig.add_subplot(grid[0, 0])
    ax_pot.set_ylabel('$V(x)$')
    ax_pot.set_xlabel('$x$')
    ax_pot.set_xlim(XLIM_POT)
    ax_pot.set_ylim(YLIM_POT)

    ax_hist = fig.add_subplot(grid[0, 1])
    ax_hist.set_xlabel('$x$')
    ax_hist.set_ylabel('No. of counts')
    ax_hist.set_xlim(XLIM)
    ax_hist.set_ylim(0.0, 1.0)

    ax_all_hist = fig.add_subplot(grid[1, 0])
    ax_all_hist.set_ylabel('Probability density')
    ax_all_hist.set_xlabel('$x$')
    ax_all_hist.set_xlim(XLIM)

    ax_hist_log = fig.add_subplot(grid[1, 1])
    ax_hist_log.set_visible(False)

    axes = {
        'pot': ax_pot,
        'hist': ax_hist,
        'hist-scaled': ax_all_hist,
        'hist-log': ax_hist_log
    }

    fig.subplots_adjust(left=0.1, bottom=0.2, right=0.90, top=0.90,
                        wspace=0.3, hspace=0.2)
    # Add some plot objects:

    plot_objects = {'figtxt': plt.figtext(0.05, 0.05, '')}
    pot_obj = add_potential_plot(ax_pot, system, forcefield)
    plot_objects.update(pot_obj)
    hist_obj = add_histogram_plots(ax_hist, ax_all_hist, ax_hist_log)
    plot_objects.update(hist_obj)
    fig.canvas.draw()
    return fig, axes, plot_objects


def update_window_plot(axes, axv, window):
    """Plot region for the given umbrella window."""
    region = np.array([[window[0], 0.0], [window[0], 1.0],
                       [window[1], 1.0], [window[1], 0.0],
                       [window[0], 0.0]])
    axv.set_xy(region)
    axes.draw_artist(axv)


def update_histogram(positions, axes, rects, store=None, density=False):
    """Update histograms according to the new data."""
    hist, bins, bin_mid = histogram(positions, bins=BINS, limits=HISTLIM,
                                    density=density)
    if density:
        ymax = max([histi[0].max() for histi in store] + [hist.max()])
        axes.set_ylim(0.0, ymax * 1.05)
    else:
        axes.set_ylim(0.0, hist.max() * 1.05)
    for rect, histi in zip(rects, hist):
        rect.set_height(histi)
        rect.set_visible(True)
    return hist, bins, bin_mid


def make_plots(system, forcefield, trajectory, energy):
    """Manage the creation of the plots."""
    fig, axes, plot_obj = create_plots(system, forcefield)
    all_histograms = plot_trials(fig, axes, plot_obj, system, forcefield,
                                 trajectory, energy)
    plot_scalings(fig, axes, plot_obj, system, forcefield, all_histograms)


def update_figure(fig, axes, artist):
    """Use blit to update the trial/accetpted position."""
    background = fig.canvas.copy_from_bbox(axes.bbox)
    fig.canvas.restore_region(background)
    axes.draw_artist(artist)
    fig.canvas.blit(axes.bbox)


def plot_trials(fig, axes, plot_obj, system, forcefield, trajectory, energy):
    """Plot the trial moves and corresponding histograms."""
    tot_step = 0  # Total number of steps done.
    all_histograms = []  # For storing all histograms.
    all_normed_histograms = []  # For storing all normed histograms.
    print_to_screen('Making plots for windows', level='success')
    for i in trange(len(UMBRELLA_WINDOWS), bar_format=BAR_FMT['window']):
        pos, trial, success = (trajectory[i][0], trajectory[i][1],
                               trajectory[i][2])
        ener = energy[i]
        update_window_plot(axes['pot'], plot_obj['window'],
                           UMBRELLA_WINDOWS[i])
        pos_so_far = []
        for j in trange(len(pos), bar_format=BAR_FMT['pos']):
            tot_step += 1
            pos_so_far.append(pos[j])

            if not success[j]:  # Add trial point:
                system.particles.pos = trial[j]
                vpot = forcefield.evaluate_potential(system)
                plot_obj['scatter_trial'].set_offsets([trial[j], vpot])
                plot_obj['scatter_trial'].set_visible(True)
            else:
                plot_obj['scatter_trial'].set_visible(False)
            if j % FIG_FREQ == 0:
                plot_obj['figtxt'].set_text(
                    f'Total number of MC cycles: {tot_step}'
                )
                plot_obj['scatter'].set_offsets([pos[j], ener[j]])
                update_histogram(pos_so_far, axes['hist'],
                                 plot_obj['counts'][i])
                update_histogram(pos_so_far, axes['hist-scaled'],
                                 plot_obj['prob'][i],
                                 store=all_normed_histograms, density=True)
                update_figure(fig, axes['pot'], plot_obj['scatter'])
                fig.savefig(f'frame-{i:03d}-{j:05d}.png')
        # Add final histograms:
        hist1 = update_histogram(pos_so_far, axes['hist'],
                                 plot_obj['counts'][i])
        all_histograms.append(hist1)
    return all_histograms


def plot_scalings(fig, axes, plot_obj, system, forcefield, all_histograms):
    """Plot the scaling for matched histograms."""
    _, scale_factors, hist_avg = match_all_histograms(
        all_histograms,
        UMBRELLA_WINDOWS,
    )
    axes['hist'].set_xlabel('$x$')
    axes['hist'].set_ylabel('Matching histograms')
    axes['hist-log'].set_xlim(XLIM)
    axes['hist-log'].set_yscale('log')
    axes['hist-log'].set_xlabel('$x$')
    axes['hist-log'].set_ylabel('Matching - logscale')
    axes['hist-log'].set_visible(True)
    plot_obj['figtxt'].set_text('Re-scaling histograms!')

    logmin = min(hist_avg[np.where(hist_avg > 0.0)[0]])

    for rects in plot_obj['counts']:
        for rec in rects:
            rec.visible = False

    for i in trange(len(UMBRELLA_WINDOWS), bar_format=BAR_FMT['window-scale']):
        hist, _, bin_mid = all_histograms[i]
        rects = plot_obj['counts'][i]
        rects_log = plot_obj['histlog'][i]
        if i == 0:
            axes['hist'].set_ylim(0.0, hist.max() * 1.05)
            axes['hist-log'].set_ylim(logmin / 10, hist.max() * 10)
            scales = np.ones(SCALE_STEPS)
        else:
            scales = np.linspace(1, scale_factors[i], SCALE_STEPS)
            plot_obj['figtxt'].set_text(f'Re-scaling histogram {i}')

        for j in trange(len(scales), bar_format=BAR_FMT['scale']):
            for rec, rec_log, histi in zip(rects, rects_log, hist):
                rec.set_height(histi * scales[j])
                rec_log.set_height(histi * scales[j])
                rec.set_visible(True)
                rec_log.set_visible(True)
            if j % 1 == 0:
                fig.savefig(f'scale-{i:03d}-{j:05d}.png')
    plot_obj['figtxt'].set_text('Done with scaling: Calculating free energy')
    axes['hist'].plot(bin_mid, hist_avg, lw=6, color='orangered', alpha=0.65)
    axes['hist-log'].plot(bin_mid, hist_avg, lw=6, color='orangered',
                          alpha=0.65)

    free = -np.log(hist_avg) / system.temperature['beta']  # free energy
    free -= free.min()
    # Replot potential:
    pos = np.linspace(-2, 2, 250)
    vpot = evaluate_potential(system, forcefield, pos)
    vpot -= vpot.min()
    axes['pot'].plot(bin_mid, free, lw=6, color='orangered', alpha=0.65)

    where = np.where(np.logical_and(pos >= XLIM_POT[0], pos <= XLIM_POT[1]))[0]

    axes['pot'].set_ylim(-0.05, vpot[where].max())
    for obj in ('scatter', 'scatter_trial', 'window'):
        plot_obj[obj].set_visible(False)
    plot_obj['potential'].set_xdata(pos)
    plot_obj['potential'].set_ydata(vpot)
    fig.savefig('final.png')


def main():
    """Run the simulation and do the plotting'."""
    trajectory, energy = run_simulation()
    # Create force field and system for helping with plotting:
    unbiased = {
        'forcefield': {
            'description': 'Double well',
        },
        'potential': [DOUBLE_WELL]
    }
    forcefield = create_force_field(unbiased)
    # Create a fake system for calculating the potential:
    system = System(
        units=DEFAULT_SETTINGS['system']['units'],
        temperature=DEFAULT_SETTINGS['system']['temperature'],
    )
    system.particles = Particles()
    make_plots(system, forcefield, trajectory, energy)


if __name__ == '__main__':
    colorama.init(autoreset=True)
    main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Module for formatting snapshot data from PyRETIS.

Important classes defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

SnapshotFormatter (:py:class:`.SnapshotFormatter`)
    Generic class for formatting system snapshots (coordinates). Note
    it is intended to format objects like :py:class:`.System`. The
    format is very similar to the XYZ format.

SnapshotFile (:py:class:`.SnapshotFile`)
    A class for a collection of snapshots.

Important methods defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

adjust_coordinate (:py:func:`.adjust_coordinate`)
    Helper method to add dimensions when formatting data in 1D or 2D
    to a format that requires 3D data.

read_txt_snapshots (:py:func:`.read_txt_snapshots`)
    For reading PyRETIS snapshots from a file.

"""
import logging
import numpy as np
from pyretis.core.common import initiate_instance
from pyretis.inout.formats.formatter import OutputFormatter
from pyretis.inout.fileio import FileIO
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())

np.set_printoptions(legacy='1.21')

__all__ = [
    'adjust_coordinate',
    'read_txt_snapshots',
    'SnapshotFormatter',
    'SnapshotFile',
]


class SnapshotFormatter(OutputFormatter):
    """Generic class for formatting system snapshots.

    Attributes
    ----------
    write_vel : boolean
        If True, we will also format velocities
    fmt : string
        Format to use for position output.
    fmt_vel : string
        Format to use for position and velocity output.

    """

    data_keys = ('atomname', 'x', 'y', 'z', 'vx', 'vy', 'vz')
    _FMT_FULL = '{} {} {} {}'
    _FMT_FULL_VEL = '{} {} {} {} {} {} {}'
    _FMT = '{:5s} {:15.9f} {:15.9f} {:15.9f}'
    _FMT_VEL = '{:5s} {:15.9f} {:15.9f} {:15.9f} {:15.9f} {:15.9f} {:15.9f}'

    def __init__(self, write_vel=True, fmt=None):
        """Initialise the formatter.

        Parameters
        ----------
        write_vel : boolean, optional
            If True, the formatter will attempt to output velocities.
        fmt : string, optional
            Selects the format to use.

        """
        super().__init__('SnapshotFormatter', header=None)
        self.print_header = False
        self.write_vel = write_vel
        if fmt == 'full':
            self.fmt = self._FMT_FULL
            self.fmt_vel = self._FMT_FULL_VEL
        else:
            self.fmt = self._FMT
            self.fmt_vel = self._FMT_VEL

    def format(self, step, data):
        """Generate the snapshot output.

        Parameters
        ----------
        step : integer
            The current step number for generating the output.
        data : object like :py:class:`.System`
            The system we are generating output for.

        """
        for lines in self.format_snapshot(step, data):
            yield lines

    def _format_without_vel(self, particles):
        """Format positions of particles for output.

        Parameters
        ----------
        particles : object like :py:class:`.Particles`
            The particles for which we will format information.

        Yields
        ------
        out : string
            The formatted output, to be written.

        """
        pos = adjust_coordinate(particles.pos)
        for namei, posi in zip(particles.name, pos):
            yield self.fmt.format(namei, *posi)

    def _format_with_vel(self, particles):
        """Format positions of particles for output.

        Parameters
        ----------
        particles : object like :py:class:`.Particles`
            The particles for which we will format information.

        Yields
        ------
        out : string
            The formatted output, to be written.

        """
        pos = adjust_coordinate(particles.pos)
        vel = adjust_coordinate(particles.vel)
        for namei, posi, veli in zip(particles.name, pos, vel):
            yield self.fmt_vel.format(namei, posi[0], posi[1], posi[2],
                                      veli[0], veli[1], veli[2])

    @staticmethod
    def parse(line):
        """Not implemented - line parser for snapshots.

        For snapshots, we use a specialized reader that will read entire
        system snapshots. To avoid confusion, we just give a warning here.

        """
        logger.warning('The line parser is not implemented for the'
                       'snapshot reader.')

    def format_snapshot(self, step, system):
        """Format the given snapshot.

        Parameters
        ----------
        step : int
            The current simulation step.
        system : object like :py:class:`.System`
            The system object with the positions to format.

        Returns
        -------
        out : list of strings
            The formatted snapshot.

        """
        npart = system.particles.npart
        buff = [
            f'{npart}',
            f'Snapshot, step: {step} box: {system.box.print_length()}',
        ]
        if self.write_vel:
            formatter = self._format_with_vel
        else:
            formatter = self._format_without_vel
        for lines in formatter(system.particles):
            buff += [lines]
        return buff

    def load(self, filename):
        """Read snapshots from a given file.

        Parameters
        ----------
        filename : string
            The path/filename to open.

        Yields
        ------
        out : dict
            This dict contains the snapshot.

        """
        for snapshot in read_txt_snapshots(filename,
                                           data_keys=self.data_keys):
            yield snapshot


def read_txt_snapshots(filename, data_keys=None):
    """Read snapshots from a text file.

    Parameters
    ----------
    filename : string
        The file to read from.
    data_keys : tuple of strings, optional
        This tuple determines the data we are to read. It can
        be of type ``('atomname', 'x', 'y', 'z', ...)``.

    Yields
    ------
    out : dict
        A dictionary with the snapshot.

    """
    lines_to_read = 0
    snapshot = None
    if data_keys is None:
        data_keys = ('atomname', 'x', 'y', 'z', 'vx', 'vy', 'vz')
    read_header = False
    with open(filename, 'r', encoding="utf8") as fileh:
        for lines in fileh:
            if read_header:
                snapshot = {'header': lines.strip()}
                snapshot['box'] = get_box_from_header(snapshot['header'])
                read_header = False
                continue
            if lines_to_read == 0:  # new snapshot
                if snapshot is not None:
                    yield snapshot
                try:
                    lines_to_read = int(lines.strip())
                except ValueError:
                    logger.error('Error in the input file %s', filename)
                    raise
                read_header = True
                snapshot = None
            else:
                lines_to_read -= 1
                data = lines.strip().split()
                for i, (val, key) in enumerate(zip(data, data_keys)):
                    if i == 0:
                        value = val.strip()
                    else:
                        value = float(val)
                    try:
                        snapshot[key].append(value)
                    except KeyError:
                        snapshot[key] = [value]
    if snapshot is not None:
        yield snapshot


def get_box_from_header(header):
    """Get box lengths from a text header.

    Parameters
    ----------
    header : string
        Header from which we will extract the box.

    Returns
    -------
    out : numpy.array or None
        The box lengths.

    """
    low = header.lower()
    if low.find('box:') != -1:
        txt = low.split('box:')[1].strip()
        return np.array([float(i) for i in txt.split()])
    return None


def adjust_coordinate(coord):
    """Adjust the dimensionality of coordinates.

    This is a helper method for trajectory writers.

    A lot of the different formats expects us to have 3 dimensional
    data. This method just adds dummy dimensions equal to zero.

    Parameters
    ----------
    coord : numpy.array
        The real coordinates.

    Returns
    -------
    out : numpy.array
        The "zero-padded" coordinates.

    """
    if len(coord.shape) == 1:
        npart, dim = len(coord), 1
    else:
        npart, dim = coord.shape
    if dim == 3:
        # correct dimensionality, just stop here:
        return coord
    adjusted = np.zeros((npart, 3))
    try:
        for i in range(dim):
            adjusted[:, i] = coord[:, i]
    except IndexError:
        if dim == 1:
            adjusted[:, 0] = coord
    return adjusted


class SnapshotFile(FileIO):
    """A class for a collection of snapshots."""

    def __init__(self, filename, file_mode, backup=True, format_settings=None):
        """Create the snapshot file with the possible settings."""
        if format_settings is not None:
            fmt = initiate_instance(SnapshotFormatter, format_settings)
        else:
            fmt = SnapshotFormatter()
        super().__init__(filename, file_mode, fmt, backup=backup)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test the restart methods."""
from io import StringIO
import logging
import os
import unittest
import tempfile
import numpy as np
from unittest.mock import patch
from pyretis.simulation.simulation import Simulation
from pyretis.core.box import create_box
from pyretis.core.system import System
from pyretis.core.common import big_fat_comparer
from pyretis.core.random_gen import create_random_generator
from pyretis.core.particles import Particles
from pyretis.forcefield.forcefield import ForceField
from pyretis.forcefield.potentials import PairLennardJonesCutnp
from pyretis.setup.createsystem import create_system
from pyretis.core.units import (units_from_settings,
                                create_conversion_factors)
from pyretis.engines.internal import VelocityVerlet
from pyretis.inout.settings import (add_default_settings,
                                    fill_up_tis_and_retis_settings,
                                    SECTIONS)
from pyretis.inout.restart import (write_restart_file,
                                   write_ensemble_restart,
                                   read_restart_file)
from pyretis.setup.createsimulation import create_retis_simulation
from pyretis.inout.common import make_dirs
from pyretis.tools.lattice import generate_lattice


logging.disable(logging.CRITICAL)

HERE = os.path.abspath(os.path.dirname(__file__))


def create_test_system():
    """Create a system we can use for testing."""
    create_conversion_factors('lj')
    xyz, size = generate_lattice('fcc', [2, 2, 2], density=0.9)
    size = np.array(size)
    box = create_box(low=size[:, 0], high=size[:, 1])
    system = System(units='lj', box=box, temperature=2.0)
    system.particles = Particles(dim=3)
    for pos in xyz:
        system.add_particle(pos, vel=np.zeros_like(pos),
                            force=np.zeros_like(pos),
                            mass=1.0, name='Ar', ptype=0)
    rgen = create_random_generator({'seed': 0})
    system.generate_velocities(rgen, distribution='maxwell', momentum=True)
    potentials = [
        PairLennardJonesCutnp(dim=3, shift=True, mixing='geometric'),
    ]
    parameters = [
        {0: {'sigma': 1, 'epsilon': 1, 'rcut': 2.5}},
    ]
    system.forcefield = ForceField(
        'Lennard Jones force field',
        potential=potentials,
        params=parameters,
    )

    return system


class TestRestartMethods(unittest.TestCase):
    """Test methods defined in the module."""

    def test_write_and_read(self):
        """Test write/read for simulation restart files."""
        settings = {
            'simulation': {},
            'system': {
                'dimensions': 3,
                'units': 'reduced',
                'temperature': 1.0
            },
            'particles': {
                'position': {
                    'generate': 'fcc',
                    'repeat': [2, 2, 2],
                    'density': 0.9,
                }
            },
            'potential': [
                {
                    'class': 'PairLennardJonesCutnp',
                    'shift': True,
                    'parameter': {
                        0: {'sigma': 1.0, 'epsilon': 1.0, 'rcut': 2.5}
                    }
                }
            ]
        }
        simulation = Simulation(settings, {'steps': 100})
        add_default_settings(settings)
        units_from_settings(settings)
        system = create_system(settings)
        simulation.system = system

        with tempfile.NamedTemporaryFile() as tmp:
            write_restart_file(tmp.name, simulation)
            tmp.flush()
            read = read_restart_file(tmp.name)
            for key in ['simulation', 'system']:
                self.assertTrue(key in read)
            self.assertTrue(big_fat_comparer(read, simulation.restart_info()))

    def test_read_write_ensemble(self):
        """Test read/write for path ensemble restart files."""
        settings = {
            'simulation': {'task': 'retis',
                           'interfaces': [-1., 0., 1.],
                           'exe_path': '.',
                           'steps': 10,
                           'zero_ensemble': True,
                           'flux': True},
            'tis': SECTIONS['tis'],
            'retis': SECTIONS['retis'],
            'engine': {'obj': VelocityVerlet(0.002)},
            'particles': {'type': 'internal'},
            'system': {'type': 'internal', 'units': 'lj',
                       'obj': create_test_system()}}

        settings['orderparameter'] = {'class': 'Position',
                                      'dim': 'x', 'index': 0,
                                      'periodic': False}
        with tempfile.TemporaryDirectory() as tempdir:
            settings['simulation']['exe_path'] = tempdir
            fill_up_tis_and_retis_settings(settings)

            with patch('sys.stdout', new=StringIO()):
                simulation = create_retis_simulation(settings)

            filename = os.path.join(tempdir, 'devil.rst')

            write_restart_file(filename, simulation)
            self.assertEqual(os.path.exists(filename), 1)

            info = simulation.restart_info()
            info_file = read_restart_file(filename)

            big_fat_comparer(info, info_file, hard=True)

            ensembles = simulation.ensembles
            for ens_set, ensemble in zip(settings['ensemble'], ensembles):
                for name in ensemble['path_ensemble'].directories():
                    make_dirs(os.path.join(tempdir, name))
                write_ensemble_restart(ensemble, ens_set)
                restart_file = os.path.join(
                    tempdir,
                    ensemble['path_ensemble'].directory['path_ensemble'],
                    'ensemble.restart')
                read = read_restart_file(restart_file)
                restart = {**ens_set,
                           **{'system': ensemble['system'].restart_info(),
                              'engine': ensemble['engine'].restart_info(),
                              'path_ensemble':
                                  ensemble['path_ensemble'].restart_info(),
                              'rgen': ensemble['rgen'].get_state(),
                              'order_function':
                              ensemble['order_function'].restart_info()
                              }
                           }

                self.assertTrue(big_fat_comparer(read, restart, hard=True))


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test the GROMACS Engine class."""
import logging
import numpy as np
import os
import shutil
import tempfile
import unittest
from pyretis.core.path import Path
from pyretis.core.common import counter
from pyretis.engines import GromacsEngine
from pyretis.inout.common import make_dirs
from pyretis.inout.formats.gromacs import (read_gromacs_gro_file,
                                           read_gromacs_generic)
from pyretis.orderparameter.orderparameter import Position
from .test_helpers.test_helpers import make_test_system

logging.disable(logging.CRITICAL)
HERE = os.path.abspath(os.path.dirname(__file__))
GMX = os.path.join(HERE, 'mockgmx.py')
MDRUN = os.path.join(HERE, 'mockmdrun.py')
GMX_DIR = os.path.join(HERE, 'gmx_input')
GMX_DIR2 = os.path.join(HERE, 'gmx_input2')
GMX_DIR3 = os.path.join(HERE, 'gmx_input3')
GMX_LOAD = os.path.join(HERE, '../initiation/gromacs/gmx_input')


class GromacsEngineTest(unittest.TestCase):
    """Run the tests for the GROMACS Engine."""

    def test_init(self):
        """Test that we can initiate the engine."""
        eng = GromacsEngine(gmx='echo',
                            mdrun='echo',
                            input_path=GMX_DIR,
                            timestep=0.002,
                            subcycles=10,
                            maxwarn=10,
                            gmx_format='gro',
                            write_vel=True,
                            write_force=False)
        eng.exe_dir = GMX_DIR
        with self.assertRaises(ValueError):
            GromacsEngine(gmx='echo',
                          mdrun='echo',
                          input_path='gmx_input',
                          timestep=0.002,
                          subcycles=10,
                          gmx_format='not-a-format')
        # Test with an index.ndx file:
        eng2 = GromacsEngine(gmx='echo',
                             mdrun='echo',
                             input_path=GMX_DIR3,
                             timestep=0.002,
                             subcycles=10,
                             maxwarn=10,
                             gmx_format='gro',
                             write_vel=True,
                             write_force=False)
        eng2.exe_dir = GMX_DIR3

    def test_read_generic(self):
        """Test a single read."""
        for i, _ in enumerate(read_gromacs_generic('no_file.g96')):
            max_i = i
        self.assertEqual(max_i, 0)
        any_g96 = os.path.join(HERE, '..', 'inout', 'config.g96')
        for i, _ in enumerate(read_gromacs_generic(any_g96)):
            max_i = i
        self.assertEqual(max_i, 0)
        any_gro = os.path.join(HERE, 'gmx_input', 'conf.gro')
        for i, _ in enumerate(read_gromacs_generic(any_gro)):
            max_i = i
        self.assertEqual(max_i, 0)
        any_trr = os.path.join(HERE, '..', 'tools', '2water.trr')
        for i, _ in enumerate(read_gromacs_generic(any_trr)):
            max_i = i
        self.assertEqual(max_i, 5)

    def test_single_step(self):
        """Test a single step using the MOCK GROMACS engine."""
        with tempfile.TemporaryDirectory() as tempdir:
            eng = GromacsEngine(gmx=GMX,
                                mdrun=MDRUN,
                                input_path=GMX_DIR,
                                timestep=0.002,
                                subcycles=3,
                                maxwarn=1,
                                gmx_format='gro',
                                write_vel=True,
                                write_force=False)
            rundir = os.path.join(tempdir, 'generategmx')
            # Create the directory for running:
            make_dirs(rundir)
            eng.exe_dir = rundir
            # Create the system:
            system = make_test_system((eng.input_files['conf'], 0))
            out = eng.step(system, 'gmx_mock_step')
            self.assertEqual(out, 'gmx_mock_step.gro')
            # Check that output files are present:
            for i in ('conf.gro', 'gmx_mock_step.gro'):
                self.assertTrue(os.path.isfile(os.path.join(rundir, i)))
            # Check that output files contain the expected data:
            _, xyz1, vel1, _ = read_gromacs_gro_file(
                os.path.join(rundir, 'conf.gro')
            )
            _, xyz2, vel2, _ = read_gromacs_gro_file(
                os.path.join(rundir, 'gmx_mock_step.gro')
            )
            self.assertTrue(np.allclose(xyz2 - xyz1,
                                        eng.subcycles * np.ones_like(xyz1)))
            self.assertTrue(np.allclose(
                vel1,
                np.repeat([0.1111, 0.2222, 0.3333], 27).reshape(3, 27).T
            ))
            self.assertTrue(np.allclose(vel2, np.ones_like(vel2)))
            # Check the final state:
            self.assertAlmostEqual(eng.subcycles * 1.0, system.particles.ekin)
            self.assertAlmostEqual(eng.subcycles * -1.0, system.particles.vpot)
            self.assertEqual(
                system.particles.get_pos()[0],
                os.path.join(rundir, 'gmx_mock_step.gro')
            )
            eng.clean_up()

    def test_modify_velocities(self):
        """Test the modify velocities method."""
        with tempfile.TemporaryDirectory() as tempdir:
            eng = GromacsEngine(gmx=GMX,
                                mdrun=MDRUN,
                                input_path=GMX_DIR,
                                timestep=0.002,
                                subcycles=10,
                                maxwarn=1,
                                gmx_format='gro',
                                write_vel=False,
                                write_force=False)
            rundir = os.path.join(tempdir, 'generategmxvel')
            # Create the directory for running:
            make_dirs(rundir)
            eng.exe_dir = rundir
            # Create the system:
            system = make_test_system((eng.input_files['conf'], 0))
            ensemble = {'system': system}
            dek, kin_new = eng.modify_velocities(ensemble,
                                                 {'aimless': True,
                                                  'momentum': False})

            self.assertAlmostEqual(kin_new, 1234.5678)
            self.assertTrue(dek == float('inf'))
            # Check that aiming fails:
            with self.assertRaises(NotImplementedError):
                eng.modify_velocities(ensemble, {'aimless': False,
                                                 'momentum': False})
            # Check that rescaling fails:
            with self.assertRaises(NotImplementedError):
                eng.modify_velocities(ensemble, {'aimless': False,
                                                 'momentum': True,
                                                 'rescale': 11})

            dek, kin_new = eng.modify_velocities(ensemble,
                                                 {'aimless': True,
                                                  'momentum': False})
            self.assertAlmostEqual(kin_new, 1234.5678)
            self.assertAlmostEqual(dek, 0.0)
            eng.clean_up()

    def test_propagate_forward(self):
        """Test the propagate method forward in time."""
        with tempfile.TemporaryDirectory() as tempdir:
            eng = GromacsEngine(gmx=GMX,
                                mdrun=MDRUN,
                                input_path=GMX_DIR,
                                timestep=0.002,
                                subcycles=7,
                                maxwarn=1,
                                gmx_format='gro',
                                write_vel=True,
                                write_force=False)
            rundir = os.path.join(tempdir, 'generategmxf')
            # Create the directory for running:
            make_dirs(rundir)
            eng.exe_dir = rundir
            # Create the system:
            system = make_test_system((eng.input_files['conf'], 0))
            # Propagate:
            orderp = Position(0, dim='x', periodic=False)
            path = Path(None, maxlen=8)
            ensemble = {'system': system,
                        'order_function': orderp,
                        'interfaces': [-0.45, 10.0, 14.0]}
            success, _ = eng.propagate(path, ensemble, reverse=False)
            initial_x = -0.422
            for i, point in enumerate(path.phasepoints):
                self.assertAlmostEqual(point.particles.ekin, eng.subcycles * i)
                self.assertAlmostEqual(point.particles.vpot,
                                       -1.0 * eng.subcycles * i)
                self.assertAlmostEqual(point.order[0],
                                       i * eng.subcycles + initial_x, places=3)
            self.assertTrue(success)
            eng.clean_up()

    def test_propagate_backward(self):
        """Test the propagate method backward in time."""
        with tempfile.TemporaryDirectory() as tempdir:
            eng = GromacsEngine(gmx=GMX,
                                mdrun=MDRUN,
                                input_path=GMX_DIR,
                                timestep=0.002,
                                subcycles=3,
                                maxwarn=1,
                                gmx_format='gro',
                                write_vel=True,
                                write_force=False)
            rundir = os.path.join(tempdir, 'generategmxb')
            # Create the directory for running:
            make_dirs(rundir)
            eng.exe_dir = rundir
            # Create the system:
            system = make_test_system((eng.input_files['conf'], 0))
            # Propagate:
            orderp = Position(0, dim='x', periodic=False)
            path = Path(None, maxlen=3)
            ensemble = {'system': system,
                        'order_function': orderp,
                        'interfaces': [-0.45, 10.0, 14.0]}
            counter.count = -1
            success, _ = eng.propagate(path, ensemble, reverse=True)
            self.assertFalse(success)
            # Check that velocities were reversed:
            _, _, vel1, _ = read_gromacs_gro_file(
                os.path.join(rundir, '0_conf.gro')
            )
            _, _, vel2, _ = read_gromacs_gro_file(
                os.path.join(rundir, 'r_0_conf.gro')
            )
            self.assertTrue(np.allclose(vel1, -1.0 * vel2))
            eng.clean_up()

    def test_integrate(self):
        """Test the integrate method."""
        with tempfile.TemporaryDirectory() as tempdir:
            eng = GromacsEngine(gmx=GMX,
                                mdrun=MDRUN,
                                input_path=GMX_DIR,
                                timestep=0.002,
                                subcycles=3,
                                maxwarn=1,
                                gmx_format='gro',
                                write_vel=True,
                                write_force=False)
            rundir = os.path.join(tempdir, 'gmxintegrate')
            # Create the directory for running:
            make_dirs(rundir)
            eng.exe_dir = rundir
            # Create the system:
            system = make_test_system((eng.input_files['conf'], 0))
            # Propagate:
            order_function = Position(0, dim='x', periodic=False)
            i = 0
            initial_x = -0.422
            ensemble = {'system': system, 'order_function': order_function}
            for step in eng.integrate(ensemble, 3):
                self.assertAlmostEqual(step['order'][0],
                                       i * eng.subcycles + initial_x,
                                       places=3)
                i += 1
            # Clean
            eng.clean_up()

    def test_select_energy_term(self):
        """Test the select_energy_term method."""
        eng = GromacsEngine(gmx=GMX,
                            mdrun=MDRUN,
                            input_path=GMX_DIR,
                            timestep=0.002,
                            subcycles=3,
                            gmx_format='gro')
        terms = 'full'
        self.assertEqual(
            str(eng.select_energy_terms(terms)),
            r"b'Potential\nKinetic-En.\nTotal-Energy\nTemperature\nPressure'"
        )
        terms = 'path'
        self.assertEqual(
            str(eng.select_energy_terms(terms)),
            r"b'Potential\nKinetic-En.'"
        )
        terms = 'non_allowed_term'  # Will be equal to terms = 'path'.
        self.assertEqual(
            str(eng.select_energy_terms(terms)),
            r"b'Potential\nKinetic-En.'"
        )

    def test_check_fails(self):
        """Test behavior if orderfunction not given."""
        eng = GromacsEngine(gmx=GMX,
                            mdrun=MDRUN,
                            input_path=GMX_DIR,
                            timestep=0.002,
                            subcycles=1,
                            gmx_format='gro')
        with tempfile.TemporaryDirectory() as tempdir:
            rundir = os.path.join(tempdir, 'gmxintegrate')
            # Create the directory for running:
            make_dirs(rundir)
            eng.exe_dir = rundir
            # Create the system:
            system = make_test_system((eng.input_files['conf'], 0))
            # Propagate:
            for step in eng.integrate({'system': system}, 1):
                self.assertTrue("thermo" in step)
                self.assertFalse("order" in step)
            order_function = Position(0, dim='x', periodic=False)
            eng.ext = "notexistingextension"
            ensemble = {'system': system, 'order_function': order_function}
            # Propagate:
            with self.assertRaises(Exception) as context:
                for _ in eng.integrate(ensemble, 1):
                    pass
            self.assertTrue("GROMACS engine does not support reading" in
                            str(context.exception))
            path = Path(None, maxlen=3)
            ensemble = {'system': system, 'order_function': order_function,
                        'interfaces':  [-1., 0.0, 1.0]}
            with self.assertRaises(Exception) as context:
                eng.propagate(path, ensemble, reverse=True)
            self.assertTrue("GROMACS engine does not support writing" in
                            str(context.exception))

            # Add a non existing .ndx file and get an error
            eng.input_files['index'] = 'ufo.ndx'
            with self.assertRaises(Exception) as context:
                eng._execute_grompp('should_be_a_grompp.mdp',
                                    'configuration.gro', 'a_label')
            self.assertTrue("(GROMACS engine) fail" in str(context.exception))

            eng.clean_up()

    def test_trjconv(self):
        """Test behavior when extracting frames."""
        # We only test the trajectory conversion here as the
        # other parts are tested elsewhere.
        eng = GromacsEngine(gmx=GMX,
                            mdrun=MDRUN,
                            input_path=GMX_DIR,
                            timestep=0.002,
                            subcycles=1,
                            gmx_format='gro')
        with tempfile.TemporaryDirectory() as tempdir:
            eng.exe_dir = tempdir
            # Use an already done TRR file:
            any_trr = os.path.join(HERE, '..', 'tools', '2water.trr')
            any_gro = os.path.join(HERE, '..', 'tools', '2water.gro')
            eng.input_files['conf'] = any_gro
            out_gro = os.path.join(tempdir, 'output.gro')
            eng._extract_frame(any_trr, 4, out_gro)
            # Check the created output file:
            self.assertTrue(os.path.isfile(out_gro))
            with open(out_gro, 'r', encoding="utf8") as infile:
                data = [i.strip().split() for i in infile]
            # Note: this is a fake output from a fake engine
            self.assertEqual(data[0][1], 'froggy')

            # This tests only the creation of the conf.g96
            shutil.copytree(GMX_LOAD, tempdir+'2')
            anytrr = os.path.join(GMX_LOAD, '../loader/023/accepted/trajB.trr')
            eng = GromacsEngine(gmx=GMX,
                                mdrun=MDRUN,
                                input_path=tempdir+'2',
                                timestep=666,
                                subcycles=1,
                                gmx_format='g96')
            out_gro = os.path.join(tempdir+'2', 'my_uncle_jonny.g96')
            # Note, we use a fake engine and it will just copy the file
            eng.exe_dir = tempdir+'2'
            eng._extract_frame(anytrr, 0, out_gro)
            self.assertTrue(os.path.isfile(out_gro))

            shutil.rmtree(os.path.join(tempdir+'2'), 'conf.gro')
            # A partial test to avoid the need to install GROMACS to run
            with self.assertRaises(RuntimeError) as err:
                GromacsEngine(gmx=GMX,
                              mdrun=MDRUN,
                              input_path=tempdir+'2',
                              timestep=0.002,
                              subcycles=1,
                              gmx_format='g96')
            self.assertTrue('GROMACS' in str(err.exception))


if __name__ == '__main__':
    unittest.main(module='test_gromacs')



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""General functions for generating reports.

This module contains some general functions for report generation. These
functions are used by the specific report generators to format the
reports.

Important methods defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

get_template (:py:func:`.get_template`)
    Returns the template for a specific output format and report type.

render_report (:py:func:`.render_report`)
    Render a report using a template and jinja2.

generate_report (:py:func:`.generate_report`)
    Generate a specific report from analysis output.
"""
import datetime
import logging
import os
# for converting rst to html and/or latex:
import docutils.core
from docutils.writers.html4css1 import Writer as HTMLWriter
from docutils.writers.html4css1 import HTMLTranslator
# for using templates
import jinja2
from pyretis import __version__ as VERSION
from pyretis.info import PROGRAM_NAME
from pyretis.inout.report.markup import latexify_number
from pyretis.inout.report.report_md import generate_report_mdflux
from pyretis.inout.report.report_path import (generate_report_retis,
                                              generate_report_retis0,
                                              generate_report_tis,
                                              generate_report_tis_path,
                                              generate_report_repptis,
                                              generate_report_pptis_path)
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


__all__ = ['get_template', 'render_report', 'generate_report']


_DATE_FMT = '%d.%m.%Y %H:%M:%S'
# File names for pre-defined templates.
# - html is done via rst (i.e. there is no html template)
# - htm is assumed to be equal to html
# - tex is assumed to be equal to latex
_TEMPLATES = {'rst': 'report_{}.rst',
              'html': 'report_{}.rst',
              'htm': 'report_{}.rst',
              'latex': 'report_{}.tex',
              'tex': 'report_{}.tex',
              'txt': 'report_{}.txt'}

_TEMPLATE_NAMES = {'md-flux': 'mdflux',
                   'retis': 'retis',
                   'retis0': 'retis0',
                   'tis': 'tis',
                   'repptis': 'repptis',
                   'pptis': 'pptis',
                   'make-tis-files': 'tis-multiple'}

_REPORT_MAP = {'md-flux': generate_report_mdflux,
               'retis': generate_report_retis,
               'retis0': generate_report_retis0,
               'make-tis-files': generate_report_tis,
               'repptis': generate_report_repptis,
               'pptis': generate_report_pptis_path,
               'tis': generate_report_tis_path}

# Table for file extensions:
_EXT = {'rst': 'rst',
        'html': 'html',
        'htm': 'htm',  # in case some people prefer it
        'latex': 'tex',
        'tex': 'tex',
        'txt': 'txt'}


def _rst_to_html(rst):
    """Convert a reStrcuturedText string to simple HTML.

    Parameters
    ----------
    rst : string
        The string to convert.

    Returns
    -------
    out : string
        A HTML document corresponding to the input reStructuredText.

    """
    htmlwriter = HTMLWriter()
    htmlwriter.translator_class = HTMLTranslator
    override = {'output_encoding': 'unicode'}
    # custom css can be added by: 'stylesheet_path': '/path/to/style.css'
    html = docutils.core.publish_string(rst, writer=htmlwriter,
                                        settings_overrides=override)
    return html


def get_template(output, report_type, template=None):
    """Return the template to use for a specified output format.

    The output is one of the defined output types, for instance 'rst'
    for restructured text or 'latex' for latex. Different report types
    will have different templates and the report types must also be
    specified here.

    Parameters
    ----------
    output : string
        This string selects the output format for the template, i.e.,
        rst, html, latex, tex.
    template : string, optional
        The full path to the template to use. If not given/found, the
        defaults in :py:const:`._TEMPLATES` will be used.
    report_type : string
        This is the type of report we are doing, e.g. TIS or MD.

    Returns
    -------
    out[0] : string
        Filename of template to use.
    out[1] : string
        The path to the template to use.

    """
    if template is None or not os.path.isfile(template):
        # Use default template, this is located in the templates dir:
        path = os.path.dirname(os.path.abspath(__file__))
        path = os.path.join(path, 'templates')
        ltype = report_type.lower()
        template = _TEMPLATES[output].format(_TEMPLATE_NAMES[ltype])
        path_to_template = os.path.join(path, template)
        if not os.path.isfile(path_to_template):
            msg = f'Could not locate template "{path_to_template}"!'
            raise ValueError(msg)
    else:
        # user specified full path to template:
        path = os.path.dirname(template)
        template = os.path.basename(template)
    return template, path


def render_report(report, output, template, path):
    """Render a report using a template and jinja2.

    The report is given as a dictionary which is used to fill in a
    template with jinja2. The template is given as a string
    (a file name) with a path to the template. The output can also be
    specified here and this is only used to convert to HTML if that is
    the desired output.

    Parameters
    ----------
    report : dict
        This dict contains the data to be reported. It is assumed that
        this dict matches the specified template.
    output : string
        This is the desired output format. Here it's used only for
        generating HTML as this is done via rst.
    template : string
        This is the template to use (the file name).
    path :  string
        This is the template file to use (its full path).

    Returns
    -------
    out[0] : string
        The generated report in the desired format.
    out[1] : string
        The file extension (i.e. file type) for the generated report.

    Note
    ----
    The parameters `template` and `path` are typically obtained by a
    call to :py:func:`.get_template`.

    """
    env = jinja2.Environment(
        block_start_string='@{%',
        block_end_string='%}@',
        variable_start_string='@{{',
        variable_end_string='}}@',
        autoescape=jinja2.select_autoescape(['html', 'htm', 'xml']),
        loader=jinja2.FileSystemLoader(path)
    )
    # pylint: disable=maybe-no-member
    render = env.get_template(template).render(report)
    # pylint: enable=maybe-no-member
    if output == 'html':
        return _rst_to_html(render), _EXT[output]
    return render, _EXT[output]


def generate_report(report_type, analysis_results, output, template=None):
    """Generate a report of a given type with the given analysis results.

    Parameters
    ----------
    report_type : string
        Selects the kind of report we want.
    analysis_results : dict
        The results from running the analysis.
    output : string
        Output format for the report.
    template : string, optional
        The full path to the template to use. If not given/found, the
        defaults in :py:const`._TEMPLATES` will be used. This is handled
        by :py:func:`.get_template`.

    Returns
    -------
    out[0] : string
        This is the generated report in the desired format
    out[1] : string
        This is an extension which can be used when writing the report
        to a file.

    """
    report = {'version': VERSION,
              'program': PROGRAM_NAME,
              'date': datetime.datetime.now().strftime(_DATE_FMT),
              'figures': {}, 'tables': {}, 'numbers': {}}
    if report_type not in _REPORT_MAP:
        raise ValueError(f'Unkown report type {report_type}')
    generator = _REPORT_MAP[report_type]
    # Check if the output is a valid format
    if output not in _TEMPLATES:
        msg = 'Format {} not defined for {} report. Defaulting to rst'
        msg = msg.format(output, report_type)
        logger.warning(msg)
        output = 'rst'
    template, path = get_template(output, report_type, template=template)
    generated = generator(analysis_results, output=output)
    report.update(generated)
    # Remove white-space from numbers:
    for key, value in report['numbers'].items():
        if isinstance(value, str):
            report['numbers'][key] = value.strip()

    if output in ('latex', 'tex', 'html', 'htm'):
        # Latexify numbers:
        for key, value in report['numbers'].items():
            if isinstance(value, str):
                report['numbers'][key] = latexify_number(value)
    return render_report(report, output, template, path)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Plot the object relation graphs."""
import matplotlib as mpl
from matplotlib import pyplot as plt
from matplotlib.patches import Rectangle


mpl.rcParams['font.family'] = 'serif'
mpl.rcParams['font.serif'] = 'DejaVu Serif'
mpl.rcParams['savefig.directory'] = ''
mpl.rcParams['savefig.format'] = 'svg'


OBJECTS = [
    {
        'title': 'System',
        'methods': ['add_particle()'],
        'attributes-ref': ['particles', 'box', 'forcefield'],
    },
    {
        'title': 'Box',
        'methods': ['pbc_wrap()'],
    },
    {
        'title': 'Particles',
        'methods': ['get_pos()', 'set_pos()'],
        'attributes': ['pos', 'vel'],
    },
    {
        'title': 'ForceField',
        'attributes-ref': ['potential'],
    },
    {
        'title': 'PathEnsemble',
        'methods': ['store_path()'],
        'attributes': ['paths'],
        'attributes-ref': ['last_path'],
    },
    {
        'title': 'Path',
        'methods': ['append()'],
        'attributes-ref': ['phasepoints'],
    },
    {
        'title': 'PotentialFunction',
        'methods': ['potential()', 'force()', 'potential_and_force()'],
    },
    {
        'title': 'System',
        'attributes-ref': ['particles', 'forcefield'],
    },
    {
        'title': 'Simulation',
        'methods': ['step()', 'run()'],
        'attributes-ref': ['system', 'engine'],
    },
    {
        'title': 'ExternalMDEngine',
        'methods': ['step()', 'propagate()']
    },
    {
        'title': 'MDEngine',
        'methods': ['integration_step()', 'propagate()'],
    },
    {
        'title': 'ParticlesExt',
        'methods': ['get_pos()', 'set_pos()'],
        'attributes': ['config', 'vel_rev'],
    },
    {
        'title': 'System',
        'attributes-ref': ['particles', 'box', 'forcefield'],
    },
    {
        'title': 'EngineBase',
        'methods': ['integration_step()', 'propagate()'],
    },
]


def draw_object(title, methods, attributes, attributes_ref):
    """Draw a simple box representing an object."""
    origin = (0, 0)

    longest = max(
        len(title)*1.15,
        max([len(i) for i in methods + attributes + attributes_ref])
    )
    print('Longest text:', longest)

    width = longest * 6

    linew = 2
    txtheight = 18

    meth_height = 10
    meth_sep = 2
    attr_height = 10
    attr_sep = 2
    attr_ref_height = 15
    attr_ref_sep = 3

    initial_sep = 0

    if methods:
        initial_sep = 1
    else:
        if attributes:
            initial_sep = 1
        else:
            initial_sep = 6

    last_sep = 0
    if attributes_ref:
        last_sep = 3
    else:
        if attributes:
            last_sep = 3
        else:
            last_sep = 3
    if attributes_ref:
        extra = 2
    else:
        extra = 0

    height = (linew + txtheight + initial_sep + last_sep + extra +
              len(methods) * meth_height +
              len(attributes) * attr_height +
              len(attributes_ref) * attr_ref_height +
              max(len(attributes) - 1, 0) * attr_sep +
              max(len(methods) - 1, 0) * meth_sep +
              max(len(attributes_ref) - 1, 0) * attr_ref_sep)

    if attributes_ref:
        last_sep = 20
    else:
        if attributes:
            last_sep = 15
        else:
            last_sep = 15

    rtitle = Rectangle(origin, width, height, color='#43ac6a')
    origin2 = (origin[0] + linew, origin[1] + linew)
    width2 = width - 2*linew
    height2 = height - txtheight - linew
    rcontents = Rectangle(origin2, width2, height2, color='white')

    ytxt = origin[1] + height - txtheight*0.5
    xtxt = origin[0] + 2

    txt_title = mpl.text.Text(xtxt, ytxt, title, fontsize=22, color='white',
                              verticalalignment='center',
                              horizontalalignment='left')

    txt_box = [txt_title]

    xtxt += linew
    ytxt = height2 - initial_sep

    for met in methods:
        txt = mpl.text.Text(xtxt, ytxt, met, fontsize=18,
                            style='normal',
                            color='#262626',
                            verticalalignment='top',
                            horizontalalignment='left')
        txt_box.append(txt)
        ytxt -= meth_height
        ytxt -= meth_sep

    for met in attributes:
        txt = mpl.text.Text(xtxt, ytxt, met, fontsize=18,
                            style='italic',
                            color='#262626',
                            verticalalignment='top',
                            horizontalalignment='left')
        txt_box.append(txt)
        ytxt -= attr_height
        ytxt -= attr_sep

    xtxt += 2.0
    ytxt -= extra
    for attr in attributes_ref:
        txt = mpl.text.Text(
            xtxt, ytxt, attr, fontsize=18,
            style='italic', color='#262626',
            verticalalignment='top', horizontalalignment='left',
            bbox={'boxstyle':
                  'roundtooth,pad=0.25',
                  'fc': '#e6e6e6',
                  'ec': '#333333'}
        )
        # Different options for the bbox:
        # {'boxstyle': 'round,pad=0.25',
        # {'boxstyle': 'round4,pad=0.25',
        # {'boxstyle': 'sawtooth,pad=0.25',
        ytxt -= attr_ref_height
        ytxt -= attr_ref_sep
        txt_box.append(txt)
    return (rtitle, rcontents), txt_box


def make_figure(obj):
    """Just plot a figure."""
    patches, txt = draw_object(
        obj['title'],
        obj.get('methods', []),
        obj.get('attributes', []),
        obj.get('attributes-ref', [])
    )

    fig = plt.figure()
    ax1 = fig.add_subplot(111, aspect='equal')
    for patch in patches:
        ax1.add_patch(patch)
    for box in txt:
        ax1.add_artist(box)
    ax1.set_xlim(-10, 150)
    ax1.set_ylim(-10, 110)
    plt.axis('off')
    plt.show()


def make_txt_figure(txts):
    """Just plot a figure."""

    fig = plt.figure()
    ax1 = fig.add_subplot(111, aspect='equal')
    xtxt = 10
    ytxt = 10
    for txt in txts:
        txtbox = mpl.text.Text(
            xtxt, ytxt, txt, fontsize=14,
            color='#ffffff',
            style='italic',
            verticalalignment='top',
            horizontalalignment='left',
            bbox={'boxstyle': 'round,pad=0.25',
                  'fc': '#34495e',
                  'ec': '#262626'}
        )
        ytxt += 20
        ax1.add_artist(txtbox)
    ax1.set_xlim(-10, 110)
    ax1.set_ylim(-10, 110)
    plt.axis('off')
    plt.show()


def main():
    """Do all plotting."""
    for obj in OBJECTS:
        make_figure(obj)

    var = ["tuple: ('file.xyz', 123)",
           "boolean: True/False",
           "numpy.array",
           r"return copy of numpy.array",
           r"return ('file.xyz', 123)"]
    make_txt_figure(var)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Compare energy output for CP2K and PyRETIS."""
import sys
import colorama
from matplotlib import pyplot as plt
import numpy as np
from pyretis.inout import print_to_screen
from pyretis.inout.settings import parse_settings_file
from pyretis.inout.formats.cp2k import (
    read_cp2k_energy,
)


plt.style.use('seaborn-v0_8-poster')


def main(energy_file, cp2k_file, plot=False):
    """Perform the test."""
    settings = parse_settings_file('cp2k.rst')
    timestep = settings['engine']['timestep']
    subcycles = settings['engine']['subcycles']

    print_to_screen(f'Reading energy file: {energy_file}',
                    level='info')
    energy = np.loadtxt(energy_file)
    print_to_screen(f'Reading CP2K energies from file: {cp2k_file}',
                    level='info')
    energy_cp2k = read_cp2k_energy(cp2k_file)

    energy_cp2k_mse = {key: val[::subcycles] for
                       key, val in energy_cp2k.items()}

    mse_ok = obtain_mses(energy, energy_cp2k_mse)

    if plot:
        print_to_screen('\nPlotting for comparison', level='message')
        plot_comparison(energy, energy_cp2k, energy_file, cp2k_file,
                        timestep*subcycles)

    if not mse_ok:
        print_to_screen('\nComparison failed!', level='error')
        sys.exit(1)


def obtain_mses(energy, energy_cp2k, tol=1.0e-5):
    """Obtain some mean squared errors."""
    pairs = ((1, 'vpot'), (2, 'ekin'), (3, 'etot'),
             (4, 'temp'))
    for pair in pairs:
        if not len(energy[:, pair[0]]) == len(energy_cp2k[pair[1]]):
            return False
        mse = ((energy[:, pair[0]] - energy_cp2k[pair[1]])**2).mean(axis=0)
        level = 'info'
        tol_ok = True
        if tol:
            tol_ok = abs(mse) < tol
            if not tol_ok:
                level = 'error'
        print_to_screen(f'MSE {pair[1]}: {mse}', level=level)
        if not tol_ok:
            return False
    return True


def plot_comparison(energy, energy_cp2k, energy_file, cp2k_file, delta_t):
    """Just plot some properties for the paths."""
    fig1 = plt.figure()
    ax11 = fig1.add_subplot(221)
    ax12 = fig1.add_subplot(222)
    ax21 = fig1.add_subplot(223)
    ax22 = fig1.add_subplot(224)
    plabel = f'PyRETIS ({energy_file})'
    glabel = f'CP2K ({cp2k_file})'
    time = np.arange(len(energy[:, 1])) * delta_t
    ax11.plot(time, energy[:, 1], lw=4, ls='-', marker='o', label=plabel)
    ax11.plot(energy_cp2k['time'], energy_cp2k['vpot'],
              lw=2, ls='--', marker='^', label=glabel)
    ax11.legend()
    ax11.set_ylabel('Potential')

    ax12.plot(time, energy[:, 2], lw=4, ls='-', marker='o')
    ax12.plot(energy_cp2k['time'], energy_cp2k['ekin'],
              lw=2, ls='--', marker='^')
    ax12.set_ylabel('Kinetic')

    ax21.plot(time, energy[:, 3], lw=4, ls='-', marker='o')
    ax21.plot(energy_cp2k['time'], energy_cp2k['etot'],
              lw=2, ls='--', marker='^')
    ax21.set_ylabel('Total')

    ax22.plot(time, energy[:, 4], lw=4, ls='-', marker='o')
    ax22.plot(energy_cp2k['time'], energy_cp2k['temp'],
              lw=2, ls='--', marker='^')
    ax22.set_ylabel('Temperature')

    fig1.tight_layout()
    plt.show()


if __name__ == '__main__':
    colorama.init(autoreset=True)
    PLOT = len(sys.argv) > 3
    main(sys.argv[1], sys.argv[2], plot=PLOT)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Simple script to compare the outcome of two simulations.
Here we compare a TIS simulation of 50 steps to known results.
"""
import tempfile
import fileinput
import os
import sys
import colorama
from pyretis.inout import print_to_screen
from pyretis.inout.settings import parse_settings_file
from pyretis.core.pathensemble import generate_ensemble_name
from pyretis.setup.createsimulation import create_ensembles


RESULTS = 'results'


def compare_files(fpath1, fpath2):
    """Compare two files."""
    print_to_screen(f'Comparing: {fpath1} {fpath2}')
    similar = True
    with open(fpath1, 'r') as file1, open(fpath2, 'r') as file2:
        for i, (linef1, linef2) in enumerate(zip(file1, file2)):
            linef1 = linef1.rstrip('\r\n')
            linef2 = linef2.rstrip('\r\n')
            if linef1 != linef2:
                print('----------------------')
                print(f'Line number: {i} ')
                print(linef1.strip())
                print(linef2.strip())
                similar = False

        if similar:
            similar = next(file1, None) is None and next(file2, None) is None

    if similar:
        print_to_screen('\t-> Files are equal!', level='success')
    else:
        print_to_screen('\t-> Files are NOT equal!', level='error')
        sys.exit(1)
    return 0


def check_path_file(ens):
    """Check that the accepted paths seem ok.

    Parameters
    ----------
    ens : object like :py:class:`.PathEnsemble`
        The path ensemble to plot for.

    Returns
    -------
    paths : dict
        Information about the paths in the ensemble.

    """
    print_to_screen(f'\nReading for {ens.ensemble_name}')
    filename = os.path.join(generate_ensemble_name(ens.ensemble_number),
                            'pathensemble.txt')
    print_to_screen(f'Reading: {filename}')
    start = ens.start_condition
    end = ('R') if ens.ensemble_number == 0 else ('R', 'L')
    something_weird = False
    with open(filename, 'r') as inputfile:
        for lines in inputfile:
            if lines.startswith('#'):
                continue
            splitline = lines.strip().split()
            status = splitline[7]
            if status != 'ACC':
                continue
            step = int(splitline[0])
            left = splitline[3]
            middle = splitline[4]
            right = splitline[5]
            length = int(splitline[6])
            mino = float(splitline[9])
            maxo = float(splitline[10])

            if length < 3:
                print_to_screen(f'Suspicious length for path {step}',
                                level='error')
                something_weird = True
            if start != left:
                print_to_screen(
                    f'Inconsistent start: {start} != {left} (step {step})',
                    level='error')
                something_weird = True
            if middle != 'M':
                print_to_screen(
                    f'Middle differ: M != {middle} (step {step})',
                    level='error')
                something_weird = True
            if right not in end:
                print_to_screen(
                    f'Inconsistent end: {right} (step {step})',
                    level='error')
                something_weird = True
            cross = [mino < interpos < maxo for interpos in ens.interfaces]
            if ens.ensemble_number == 0:
                idx1, idx2 = 1, 2
            else:
                idx1, idx2 = 0, 1
            if not cross[idx1] or not cross[idx2]:
                something_weird = True
                print_to_screen(
                    'Inconsistent crossings: {} {} (step {})'.format(
                        cross[idx1],
                        cross[idx2],
                        step
                    ),
                    level='error'
                )
    if not something_weird:
        print_to_screen('Accepted paths are OK!', level='success')
    else:
        sys.exit(1)
    return 0


def run_check_path_file(settings):
    """Check paths using given simulation settings."""
    ensembles = create_ensembles(settings)
    retval = [check_path_file(ens['path_ensemble']) for ens in ensembles]
    return sum(retval)


def compare_path_files(settings):
    """Compare pathensemble.txt files."""
    inter = settings['simulation']['interfaces']
    retval = 0
    for i in range(1, len(inter)):
        ens_dir = generate_ensemble_name(i)
        fil1 = os.path.join(ens_dir, 'pathensemble.txt')
        fil2 = os.path.join(RESULTS, ens_dir, 'pathensemble.txt')
        ret = compare_files(fil1, fil2)
        retval += ret
    return retval


def compare_rst_files(settings):
    """Compare tis.rst files."""
    inter = settings['simulation']['interfaces']
    retval = 0
    for i in range(1, len(inter)):
        fil1 = os.path.join(f"tis-00{i}.rst")
        for line in fileinput.input(fil1, inplace=1):
            if 'exe_path' in line.split():
                line = ''
            sys.stdout.write(line)
        fil2 = os.path.join(RESULTS, f"tis-00{i}.rst")
        for line in fileinput.input(fil2, inplace=1):
            if 'exe_path' in line.split():
                line = ''
            sys.stdout.write(line)
        ret = compare_files(fil1, fil2)
        retval += ret
    return retval


def compare_prob_files():
    """Compare probabilities files."""
    names = ['tis-multiple_report.html',
             'tis-multiple_report.rst',
             'tis-multiple_report.tex']
    # Skip the lines that contain the time when the command was called
    skips = [lambda i: i in [701, 702],
             lambda i: i in [4, 5],
             lambda i: i in [24, 25]]
    retval = 0
    for name, skip in zip(names, skips):
        fil1 = os.path.join('report', name)
        fil2 = os.path.join(RESULTS, name)
        basename, ext = os.path.splitext(name)
        basename += '_'
        with tempfile.NamedTemporaryFile(prefix=basename, suffix=ext) as tmp:
            with open(tmp.name, 'wb') as outfile:
                with open(fil1, 'rb') as infile:  # This one fail?
                    for i, line in enumerate(infile):
                        if not skip(i):
                            outfile.write(line)
            ret = compare_files(tmp.name, fil2)
            retval += ret
    return retval


def main():
    """Run the full comparison."""
    sets = parse_settings_file('tis-multiple.rst')
    print_to_screen('\nCheck crossing probabilities', level='message')
    print_to_screen('============================', level='message')
    ret4 = compare_prob_files()
    print_to_screen('\nComparing tis.rst files', level='message')
    print_to_screen('=======================', level='message')
    ret1 = compare_rst_files(sets)
    print_to_screen('\nComparing pathensemble.txt files', level='message')
    print_to_screen('================================', level='message')
    ret2 = compare_path_files(sets)
    print_to_screen('\nCheck accepted paths', level='message')
    print_to_screen('====================', level='message')
    ret3 = run_check_path_file(sets)
    return ret1 + ret2 + ret3 + ret4


if __name__ == '__main__':
    colorama.init(autoreset=True)
    sys.exit(main())
